"""Export a DependencyGraph to JSON or DOT (Graphviz) format."""
# pylint: disable=duplicate-code

from __future__ import annotations

import json
import os
import re
import subprocess  # nosec B404
from pathlib import Path
from typing import Protocol, TypedDict, cast

from import_cruiser.graph import Dependency, DependencyGraph, Module, detect_cycles

JSONDict = dict[str, object]

DB_EXTERNAL_MODULES: set[str] = {
    "sqlalchemy",
    "sqlmodel",
    "alembic",
    "psycopg",
    "psycopg2",
    "asyncpg",
    "pg8000",
    "aiopg",
    "databases",
    "postgres",
}


class ViolationLike(Protocol):
    source: str
    target: str
    severity: str
    message: str
    rule_name: str

    def to_dict(self) -> dict[str, object]: ...


class Cluster(TypedDict):
    id: str
    label: str
    parent: str | None
    modules: list[Module]


def export_json(
    graph: DependencyGraph,
    violations: list[ViolationLike] | None = None,
    cycles: list[list[str]] | None = None,
) -> str:
    """Serialize *graph* (and optional analysis results) to a JSON string."""
    if violations is None:
        violations = []
    if cycles is None:
        cycles = detect_cycles(graph)

    data = cast(
        JSONDict,
        {
            "summary": {
                "modules": len(graph.modules),
                "dependencies": len(graph.dependencies),
                "cycles": len(cycles),
                "violations": len(violations),
            },
            "modules": [
                {
                    "name": m.name,
                    "path": m.path,
                    "imports": graph.dependencies_of(m.name),
                }
                for m in sorted(graph.modules, key=lambda m: m.name)
            ],
            "dependencies": [
                {"source": d.source, "target": d.target, "line": d.line}
                for d in sorted(graph.dependencies, key=lambda d: (d.source, d.target))
            ],
            "cycles": [cycle for cycle in cycles],
            "violations": [v.to_dict() for v in violations],
        },
    )
    return json.dumps(data, indent=2)


def export_dot(
    graph: DependencyGraph,
    graph_name: str = "import_cruiser",
    violations: list[ViolationLike] | None = None,
    rankdir: str = "LR",
    cluster_depth: int = 2,
    cluster_mode: str = "path",
    style: str = "depcruise",
    edge_mode: str = "node",
) -> str:
    """Return a DOT-format string representing the dependency graph."""
    if violations is None:
        violations = []
    cycles = detect_cycles(graph)
    cycle_edges = _edges_in_cycles(cycles)
    cycle_nodes = {name for cycle in cycles for name in cycle}
    violation_edges = _edges_from_violations(violations)

    graph_attrs, node_attrs, edge_attrs = _style_attrs(style, rankdir)
    depcruise = style == "depcruise"
    path_root = _common_root(graph.modules) if cluster_mode == "path" else None
    http_nodes = {
        module.name
        for module in graph.modules
        if _is_http_external_node(module.name, module.path)
    }
    external_anchor_parts = _external_anchor_parts(graph, path_root)
    node_id_map: dict[str, str] = {}
    lines = _init_dot_lines(
        graph_name=graph_name,
        graph_attrs=graph_attrs,
        node_attrs=node_attrs,
        edge_attrs=edge_attrs,
        depcruise=depcruise,
    )

    node_to_cluster: dict[str, str] = {}
    cluster_index: dict[str, list[str]] = {}
    if depcruise:
        _append_depcruise_nodes(
            lines=lines,
            modules=graph.modules,
            node_id_map=node_id_map,
            path_root=path_root,
            external_anchor_parts=external_anchor_parts,
        )
    else:
        node_to_cluster, cluster_index = _append_standard_nodes(
            lines=lines,
            modules=graph.modules,
            cycle_nodes=cycle_nodes,
            cluster_depth=cluster_depth,
            cluster_mode=cluster_mode,
            style=style,
            path_root=path_root,
        )

    if not depcruise and edge_mode == "cluster" and cluster_depth > 0:
        edge_mode = _append_cluster_edges(
            lines=lines,
            dependencies=graph.dependencies,
            node_to_cluster=node_to_cluster,
            cluster_index=cluster_index,
            edge_mode=edge_mode,
        )

    if edge_mode != "cluster" or depcruise:
        _append_dependency_edges(
            lines=lines,
            dependencies=graph.dependencies,
            node_id_map=node_id_map,
            violation_edges=violation_edges,
            cycle_edges=cycle_edges,
            depcruise=depcruise,
            http_nodes=http_nodes,
        )

    lines.append("}")
    return "\n".join(lines)


def _init_dot_lines(
    graph_name: str,
    graph_attrs: str,
    node_attrs: str,
    edge_attrs: str,
    depcruise: bool,
) -> list[str]:
    if depcruise:
        return [
            f'strict digraph "{graph_name}"{{',
            f"    {graph_attrs}",
            f"    node [{node_attrs}]",
            f"    edge [{edge_attrs}]",
            "",
        ]
    return [
        f'digraph "{graph_name}" {{',
        f"    graph [{graph_attrs}];",
        f"    node [{node_attrs}];",
        f"    edge [{edge_attrs}];",
        "",
    ]


def _append_depcruise_nodes(
    lines: list[str],
    modules: list[Module],
    node_id_map: dict[str, str],
    path_root: str | None,
    external_anchor_parts: dict[str, list[str]],
) -> None:
    for module in sorted(modules, key=lambda m: m.name):
        node_id = _depcruise_node_id(
            module,
            path_root,
            external_anchor_parts=external_anchor_parts,
        )
        node_id_map[module.name] = node_id
        lines.append(
            _depcruise_cluster_line(
                module,
                node_id,
                path_root,
                external_anchor_parts=external_anchor_parts,
            )
        )
    if modules:
        lines.append("")


def _append_standard_nodes(
    lines: list[str],
    modules: list[Module],
    cycle_nodes: set[str],
    cluster_depth: int,
    cluster_mode: str,
    style: str,
    path_root: str | None,
) -> tuple[dict[str, str], dict[str, list[str]]]:
    cluster_index: dict[str, list[str]] = {}
    node_to_cluster: dict[str, str] = {}
    root_modules: list[Module]

    if cluster_depth > 0:
        root_clusters, root_modules, flat_clusters = _build_clusters(
            modules, cluster_depth, cluster_mode
        )
        non_empty_clusters = _non_empty_clusters(flat_clusters)
        lines.extend(
            _render_cluster_tree(
                root_clusters,
                flat_clusters,
                cycle_nodes,
                indent="    ",
                mode=cluster_mode,
                allowed=non_empty_clusters,
                style=style,
            )
        )
        for module in modules:
            cluster_key = _node_cluster_key(
                module.name,
                module.path,
                cluster_mode,
                cluster_depth,
                path_root,
            )
            if cluster_key and cluster_key in non_empty_clusters:
                node_to_cluster[module.name] = cluster_key
                cluster_index.setdefault(cluster_key, []).append(module.name)
    else:
        root_modules = sorted(modules, key=lambda m: m.name)

    for module in root_modules:
        lines.extend(
            _dot_node_lines(
                module.name,
                module.path,
                module.name in cycle_nodes,
                indent="    ",
                label=_leaf_label(module, cluster_mode),
            )
        )

    if root_modules:
        lines.append("")

    return node_to_cluster, cluster_index


def _append_cluster_edges(
    lines: list[str],
    dependencies: list[Dependency],
    node_to_cluster: dict[str, str],
    cluster_index: dict[str, list[str]],
    edge_mode: str,
) -> str:
    cluster_edges: set[tuple[str, str]] = set()
    for dep in dependencies:
        src_cluster = node_to_cluster.get(dep.source)
        tgt_cluster = node_to_cluster.get(dep.target)
        if not src_cluster or not tgt_cluster or src_cluster == tgt_cluster:
            continue
        if _clusters_related(src_cluster, tgt_cluster):
            continue
        cluster_edges.add((src_cluster, tgt_cluster))

    if not cluster_edges:
        return "node"

    for src_cluster, tgt_cluster in sorted(cluster_edges):
        src_node = cluster_index[src_cluster][0]
        tgt_node = cluster_index[tgt_cluster][0]
        lines.append(
            f"    {_dot_id(src_node)} -> {_dot_id(tgt_node)} "
            f'[ltail="cluster_{_cluster_id(src_cluster)}", '
            f'lhead="cluster_{_cluster_id(tgt_cluster)}"];'
        )
    return edge_mode


def _append_dependency_edges(
    lines: list[str],
    dependencies: list[Dependency],
    node_id_map: dict[str, str],
    violation_edges: dict[tuple[str, str], ViolationLike],
    cycle_edges: set[tuple[str, str]],
    depcruise: bool,
    http_nodes: set[str],
) -> None:
    for dep in sorted(dependencies, key=lambda d: (d.source, d.target)):
        src = _dot_id(node_id_map.get(dep.source, dep.source))
        tgt = _dot_id(node_id_map.get(dep.target, dep.target))
        violation = violation_edges.get((dep.source, dep.target))
        if violation and not depcruise:
            color = _severity_color(violation.severity)
            lines.append(f'    {src} -> {tgt} [color="{color}", penwidth=2.2];')
        elif (dep.source, dep.target) in cycle_edges and not depcruise:
            lines.append(f'    {src} -> {tgt} [color="#C0392B", penwidth=1.6];')
        elif depcruise and dep.target in http_nodes:
            lines.append(
                f'    {src} -> {tgt} [color="#1D4ED8", style="dashed", '
                "penwidth=1.4, arrowsize=0.7]"
            )
        else:
            suffix = "" if depcruise else ";"
            lines.append(f"    {src} -> {tgt}{suffix}")


def export_svg(
    graph: DependencyGraph,
    graph_name: str = "import_cruiser",
    violations: list[ViolationLike] | None = None,
    engine: str = "dot",
    rankdir: str = "LR",
    cluster_depth: int = 2,
    cluster_mode: str = "path",
    style: str = "depcruise",
    edge_mode: str = "node",
) -> str:
    dot = export_dot(
        graph,
        graph_name=graph_name,
        violations=violations,
        rankdir=rankdir,
        cluster_depth=cluster_depth,
        cluster_mode=cluster_mode,
        style=style,
        edge_mode=edge_mode,
    )
    svg = _render_with_edge_fallback(
        dot=dot,
        graph=graph,
        fmt="svg",
        engine=engine,
        graph_name=graph_name,
        violations=violations,
        rankdir=rankdir,
        cluster_depth=cluster_depth,
        cluster_mode=cluster_mode,
        style=style,
        edge_mode=edge_mode,
    )
    return _add_svg_padding(svg)


def export_html(
    graph: DependencyGraph,
    graph_name: str = "import_cruiser",
    violations: list[ViolationLike] | None = None,
    engine: str = "dot",
    rankdir: str = "LR",
    cluster_depth: int = 2,
    cluster_mode: str = "path",
    style: str = "depcruise",
    edge_mode: str = "node",
) -> str:
    dot = export_dot(
        graph,
        graph_name=graph_name,
        violations=violations,
        rankdir=rankdir,
        cluster_depth=cluster_depth,
        cluster_mode=cluster_mode,
        style=style,
        edge_mode=edge_mode,
    )
    try:
        svg = _add_svg_padding(
            _render_with_edge_fallback(
                dot=dot,
                graph=graph,
                fmt="svg",
                engine=engine,
                graph_name=graph_name,
                violations=violations,
                rankdir=rankdir,
                cluster_depth=cluster_depth,
                cluster_mode=cluster_mode,
                style=style,
                edge_mode=edge_mode,
            )
        )
        body = _html_with_svg(svg, graph_name)
    except RuntimeError as exc:
        body = _html_with_fallback(dot, graph_name, str(exc))
    return body


def _render_with_edge_fallback(
    *,
    dot: str,
    graph: DependencyGraph,
    fmt: str,
    engine: str,
    graph_name: str,
    violations: list[ViolationLike] | None,
    rankdir: str,
    cluster_depth: int,
    cluster_mode: str,
    style: str,
    edge_mode: str,
) -> str:
    try:
        return _render_dot(dot, fmt, engine=engine)
    except RuntimeError as exc:
        if edge_mode != "cluster":
            raise
        cluster_safe_dot = export_dot(
            graph,
            graph_name=graph_name,
            violations=violations,
            rankdir=rankdir,
            cluster_depth=cluster_depth,
            cluster_mode=cluster_mode,
            style=style,
            edge_mode="node",
        )
        try:
            return _render_dot(cluster_safe_dot, fmt, engine=engine)
        except RuntimeError:
            raise exc


def _dot_id(name: str) -> str:
    """Convert a dotted module name to a valid DOT identifier."""
    return '"' + name.replace('"', '\\"') + '"'


def _depcruise_node_id(
    module: Module,
    root: str | None,
    external_anchor_parts: dict[str, list[str]] | None = None,
) -> str:
    if not module.path:
        anchor = (external_anchor_parts or {}).get(module.name, [])
        if anchor:
            return "/".join([*anchor, module.name])
        return module.name
    try:
        path = Path(module.path).resolve()
        if root:
            path = path.relative_to(root)
        return str(path).replace("\\", "/")
    except ValueError:
        return module.path.replace("\\", "/")


def _depcruise_cluster_line(
    module: Module,
    node_id: str,
    root: str | None,
    external_anchor_parts: dict[str, list[str]] | None = None,
) -> str:
    rel_path = node_id if module.path else module.name
    label = Path(module.path).name if module.path else module.name
    attrs = _depcruise_node_attrs(module, label, rel_path)
    parts: list[str] = []
    if module.path:
        try:
            path = Path(module.path).resolve()
            if root:
                path = path.relative_to(root)
            parts = list(path.parts[:-1])
        except ValueError:
            parts = list(Path(module.path).parts[:-1])
    elif external_anchor_parts:
        parts = external_anchor_parts.get(module.name, [])

    if not parts:
        return f"    {_dot_id(node_id)} [{attrs} ]"

    line = f'    subgraph "cluster_{parts[0]}" {{label="{parts[0]}" '
    prefix = parts[0]
    for part in parts[1:]:
        prefix = f"{prefix}/{part}"
        line += f'subgraph "cluster_{prefix}" {{label="{part}" '

    line += f"{_dot_id(node_id)} [{attrs} ] "
    line += "}" * len(parts)
    return line


def _depcruise_node_attrs(module: Module, label: str, rel_path: str) -> str:
    base = f'label=<{label}> tooltip="{label}" URL="{rel_path}"'
    if _is_database_external_node(module.name, module.path):
        return (
            f'{base} shape="cylinder" style="filled" '
            'fillcolor="#FFE7CC" color="#B45309" penwidth="1.4" '
            'fontcolor="#7C2D12"'
        )
    if _is_http_external_node(module.name, module.path):
        return (
            f'{base} shape="box" style="rounded,filled,dashed" '
            'fillcolor="#DBEAFE" color="#1D4ED8" penwidth="1.4" '
            'fontcolor="#1E3A8A"'
        )
    return base


def _dot_node_lines(
    name: str,
    path: str,
    in_cycle: bool,
    indent: str,
    label: str,
) -> list[str]:
    safe = _dot_id(name)
    attrs = [f'label="{label}"', f'tooltip="{path}"']
    if in_cycle:
        attrs.append('fillcolor="#FDECEA"')
        attrs.append('color="#C0392B"')
    return [f"{indent}{safe} [{', '.join(attrs)}];"]


def _edges_in_cycles(cycles: list[list[str]]) -> set[tuple[str, str]]:
    edges: set[tuple[str, str]] = set()
    for cycle in cycles:
        if len(cycle) < 2:
            continue
        for idx in range(len(cycle)):
            src = cycle[idx]
            tgt = cycle[(idx + 1) % len(cycle)]
            edges.add((src, tgt))
    return edges


def _build_clusters(
    modules: list[Module], depth: int, mode: str
) -> tuple[dict[str, Cluster], list[Module], dict[str, Cluster]]:
    clusters: dict[str, Cluster] = {}
    root_modules: list[Module] = []
    root_path = _common_root(modules) if mode == "path" else None

    for module in modules:
        parts = _cluster_parts(module, mode, root_path)
        if len(parts) <= 1 or depth <= 0:
            root_modules.append(module)
            continue

        max_depth = min(depth, len(parts) - (0 if mode == "path" else 1))
        parent_id = None
        for i in range(1, max_depth + 1):
            cid = ".".join(parts[:i])
            if cid not in clusters:
                clusters[cid] = {
                    "id": cid,
                    "label": parts[i - 1],
                    "parent": parent_id,
                    "modules": [],
                }
            parent_id = cid

        if parent_id is None:  # pragma: no cover
            continue
        clusters[parent_id]["modules"].append(module)

    root_clusters: dict[str, Cluster] = {}
    for cid, cluster in clusters.items():
        parent = cluster["parent"]
        if parent is None:
            root_clusters[cid] = cluster

    return root_clusters, sorted(root_modules, key=lambda m: m.name), clusters


def _render_cluster_tree(
    root_clusters: dict[str, Cluster],
    flat_clusters: dict[str, Cluster],
    cycle_nodes: set[str],
    indent: str,
    mode: str,
    allowed: set[str],
    style: str,
) -> list[str]:
    lines: list[str] = []

    def children_of(parent_id: str) -> list[Cluster]:
        return [c for c in flat_clusters.values() if c["parent"] == parent_id]

    def render_cluster(cluster: Cluster, level_indent: str) -> None:
        if cluster["id"] not in allowed:
            return
        cid = _cluster_id(cluster["id"])
        lines.append(f'{level_indent}subgraph "cluster_{cid}" {{')
        lines.append(f'{level_indent}    label="{cluster["label"]}";')
        if style != "depcruise":
            lines.append(f"{level_indent}    margin=6;")
            lines.append(f'{level_indent}    color="black";')
            lines.append(f"{level_indent}    penwidth=1.0;")
            lines.append(f'{level_indent}    style="rounded,bold";')
            lines.append(f'{level_indent}    fontname="Helvetica";')
            lines.append(f"{level_indent}    fontsize=9;")

        allowed_children = [
            child for child in children_of(cluster["id"]) if child["id"] in allowed
        ]
        if style != "depcruise" and not cluster["modules"] and allowed_children:
            anchor_id = _dot_id(f"__cluster_anchor_{cid}")
            lines.append(
                f"{level_indent}    {anchor_id} "
                '[label="", shape=point, width=0, height=0, style=invis];'
            )

        for child in sorted(allowed_children, key=lambda c: c["label"]):
            render_cluster(child, level_indent + "    ")

        for module in sorted(cluster["modules"], key=lambda m: m.name):
            lines.extend(
                _dot_node_lines(
                    module.name,
                    module.path,
                    module.name in cycle_nodes,
                    indent=level_indent + "    ",
                    label=_leaf_label(module, mode),
                )
            )

        lines.append(f"{level_indent}}}")

    for cluster in sorted(root_clusters.values(), key=lambda c: c["label"]):
        render_cluster(cluster, indent)
        lines.append("")

    return lines


def _cluster_id(group: str) -> str:
    return group.replace('"', "").replace(".", "_").replace("/", "_").replace("\\", "_")


def _non_empty_clusters(flat_clusters: dict[str, Cluster]) -> set[str]:
    children_map: dict[str, list[Cluster]] = {}
    for cluster in flat_clusters.values():
        parent = cluster["parent"]
        if parent is not None:
            children_map.setdefault(parent, []).append(cluster)

    cache: dict[str, bool] = {}

    def has_content(cluster: Cluster) -> bool:
        cid = cluster["id"]
        if cid in cache:
            return cache[cid]
        if cluster["modules"]:
            cache[cid] = True
            return True
        for child in children_map.get(cid, []):
            if has_content(child):
                cache[cid] = True
                return True
        cache[cid] = False
        return False

    return {cid for cid, cluster in flat_clusters.items() if has_content(cluster)}


def _clusters_related(a: str, b: str) -> bool:
    return a.startswith(b + ".") or b.startswith(a + ".")


def _leaf_label(module, mode: str) -> str:
    if mode == "path":
        return Path(module.path).name
    return module.name.split(".")[-1]


def _node_cluster_key(
    name: str,
    path: str,
    mode: str,
    depth: int,
    root_path: str | None,
) -> str:
    if depth <= 0:
        return ""
    if mode == "path":
        if not path or not root_path:
            return ""
        try:
            rel = Path(path).resolve().relative_to(root_path)
        except ValueError:
            return ""
        parts = list(rel.parts[:-1])
        if not parts:
            return ""
        return ".".join(parts[: min(depth, len(parts))])
    parts = name.split(".")
    if len(parts) <= 1:
        return ""
    return ".".join(parts[: min(depth, len(parts))])


def _cluster_parts(module: Module, mode: str, root_path: str | None) -> list[str]:
    if mode == "path":
        if not root_path:
            return [module.name]
        rel = Path(module.path).resolve().relative_to(root_path)
        return list(rel.parts[:-1])  # directories only
    return module.name.split(".")


def _common_root(modules: list[Module]) -> str | None:
    paths = [m.path for m in modules if m.path]
    if not paths:
        return None
    try:
        common = Path(os.path.commonpath(paths)).resolve()
        common_dir = common if common.is_dir() else common.parent
        if common_dir.name == "src":
            return str(common_dir.parent)
        if common_dir.parent.name == "src":
            return str(common_dir.parent.parent)
        return str(common_dir)
    except ValueError:
        return None


def _edges_from_violations(
    violations: list[ViolationLike],
) -> dict[tuple[str, str], ViolationLike]:
    return {(v.source, v.target): v for v in violations}


def _severity_color(severity: str) -> str:
    return {
        "error": "#C0392B",
        "warn": "#E67E22",
        "info": "#2980B9",
    }.get(severity, "#C0392B")


def _style_attrs(style: str, rankdir: str) -> tuple[str, str, str]:
    if style == "depcruise":
        graph_attrs = (
            f'rankdir="{rankdir}" splines="true" overlap="false" nodesep="0.16" '
            'ranksep="0.18" fontname="Helvetica-bold" fontsize="9" '
            'style="rounded,bold,filled" fillcolor="#ffffff" compound="true"'
        )
        node_attrs = (
            'shape="box" style="rounded, filled" height="0.2" color="black" '
            'fillcolor="#ffffcc" fontcolor="black" fontname="Helvetica" fontsize="9"'
        )
        edge_attrs = (
            'arrowhead="normal" arrowsize="0.6" penwidth="2.0" '
            'color="#0000001f" fontname="Helvetica" fontsize="9"'
        )
        return graph_attrs, node_attrs, edge_attrs

    if style == "archi":
        graph_attrs = (
            f"rankdir={rankdir}, splines=ortho, overlap=false, ranksep=0.55, "
            'nodesep=0.28, pack=true, packmode="array_u", newrank=true, '
            'compound=true, ratio="compress", '
            'fontname="Helvetica", fontsize=10, bgcolor="white", pad=0.18, margin=0.18'
        )
        node_attrs = (
            'shape=box, style="rounded,filled", fontname="Helvetica", fontsize=10, '
            'color="#2F2F2F", fillcolor="#DFF3F8"'
        )
        edge_attrs = 'color="#A0A0A0", arrowsize=0.7'
        return graph_attrs, node_attrs, edge_attrs

    if style == "cruiser":
        graph_attrs = (
            f"rankdir={rankdir}, splines=curved, overlap=false, ranksep=0.5, "
            'nodesep=0.25, pack=true, packmode="clust", newrank=true, '
            'compound=true, ratio="compress", '
            'fontname="Helvetica", fontsize=10, bgcolor="white", pad=0.18, margin=0.18'
        )
        node_attrs = (
            'shape=box, style="rounded,filled", fontname="Helvetica", fontsize=10, '
            'color="#2F2F2F", fillcolor="#DFF3F8"'
        )
        edge_attrs = 'color="#B0B0B0", arrowsize=0.7'
        return graph_attrs, node_attrs, edge_attrs

    if style == "navigator":
        graph_attrs = (
            f"rankdir={rankdir}, splines=curved, overlap=prism, ranksep=0.55, "
            'nodesep=0.32, pack=true, packmode="clust", newrank=true, '
            'compound=true, ratio="compress", '
            'fontname="Helvetica", fontsize=10, bgcolor="white", pad=0.18, margin=0.18'
        )
        node_attrs = (
            'shape=box, style="rounded,filled", fontname="Helvetica", fontsize=10, '
            'color="#2F2F2F", fillcolor="#DFF3F8"'
        )
        edge_attrs = 'color="#B0B0B0", arrowsize=0.7'
        return graph_attrs, node_attrs, edge_attrs

    graph_attrs = (
        f"rankdir={rankdir}, splines=curved, overlap=false, nodesep=0.12, ranksep=0.12, "
        'fontname="Helvetica-bold", fontsize=9, style="rounded,bold,filled", '
        'fillcolor="#FFFFFF", compound=true, pack=true, packmode="array_u", '
        'newrank=true, ratio="compress", pad=0.18, margin=0.18'
    )
    node_attrs = (
        'shape=box, style="rounded,filled", height=0.2, color="black", '
        'fillcolor="#FFFFCC", fontcolor="black", fontname="Helvetica", fontsize=9'
    )
    edge_attrs = (
        'arrowhead="normal", arrowsize=0.6, penwidth=2.0, '
        'color="#0000001f", fontname="Helvetica", fontsize=9'
    )
    return graph_attrs, node_attrs, edge_attrs


def _is_database_external_node(name: str, path: str) -> bool:
    if path:
        return False
    return name.split(".", 1)[0] in DB_EXTERNAL_MODULES


def _is_http_external_node(name: str, path: str) -> bool:
    if path:
        return False
    if "/" in name or ":" in name:
        return False
    if "." not in name:
        return False
    root = name.split(".", 1)[0]
    return root not in DB_EXTERNAL_MODULES


def _external_anchor_parts(
    graph: DependencyGraph,
    path_root: str | None,
) -> dict[str, list[str]]:
    module_map = {module.name: module for module in graph.modules}
    source_parts_by_external: dict[str, list[list[str]]] = {}
    for dep in graph.dependencies:
        target_module = module_map.get(dep.target)
        source_module = module_map.get(dep.source)
        if target_module is None or source_module is None:
            continue
        if target_module.path:
            continue
        if not source_module.path:
            continue
        parts = _module_parent_parts(source_module.path, path_root)
        if not parts:
            continue
        source_parts_by_external.setdefault(target_module.name, []).append(parts)

    return {
        external: _common_prefix(parts_list)
        for external, parts_list in source_parts_by_external.items()
    }


def _module_parent_parts(path: str, path_root: str | None) -> list[str]:
    try:
        parent = Path(path).resolve().parent
        if path_root:
            parent = parent.relative_to(Path(path_root).resolve())
        return list(parent.parts)
    except ValueError:
        return list(Path(path).resolve().parent.parts)


def _common_prefix(parts_list: list[list[str]]) -> list[str]:
    if not parts_list:
        return []
    prefix = parts_list[0][:]
    for parts in parts_list[1:]:
        idx = 0
        max_len = min(len(prefix), len(parts))
        while idx < max_len and prefix[idx] == parts[idx]:
            idx += 1
        prefix = prefix[:idx]
        if not prefix:
            break
    return prefix


def _render_dot(dot: str, fmt: str, engine: str = "dot") -> str:
    try:
        result = subprocess.run(  # nosec B603
            [engine, f"-T{fmt}"],
            input=dot,
            text=True,
            capture_output=True,
            check=True,
        )
    except FileNotFoundError as exc:
        raise RuntimeError("Graphviz 'dot' is not installed.") from exc
    except subprocess.CalledProcessError as exc:
        raise RuntimeError(exc.stderr.strip() or "Graphviz rendering failed.") from exc
    return result.stdout


def _add_svg_padding(svg: str, padding: int = 8) -> str:
    viewbox_match = re.search(
        r'viewBox="([\-0-9.eE]+)\s+([\-0-9.eE]+)\s+([\-0-9.eE]+)\s+([\-0-9.eE]+)"',
        svg,
    )
    if not viewbox_match:
        return svg

    min_x = float(viewbox_match.group(1))
    min_y = float(viewbox_match.group(2))
    width = float(viewbox_match.group(3))
    height = float(viewbox_match.group(4))

    new_viewbox = (
        f'viewBox="{min_x - padding:.2f} {min_y - padding:.2f} '
        f'{width + 2 * padding:.2f} {height + 2 * padding:.2f}"'
    )
    svg = svg[: viewbox_match.start()] + new_viewbox + svg[viewbox_match.end() :]

    def _bump_dimension(attr: str, text: str) -> str:
        match = re.search(rf'{attr}="([0-9.]+)([a-z%]*)"', text)
        if not match:
            return text
        old_value = float(match.group(1))
        unit = match.group(2)
        new_attr = f'{attr}="{old_value + 2 * padding:.0f}{unit}"'
        return text[: match.start()] + new_attr + text[match.end() :]

    svg = _bump_dimension("width", svg)
    svg = _bump_dimension("height", svg)
    return svg


def _html_with_svg(svg: str, title: str) -> str:
    display_title = _display_graph_title(title)
    return f"""<!doctype html>
<html lang="en">
<head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <title>{display_title}</title>
    <style>
        :root {{
            color-scheme: only light;
            --header-h: 44px;
            --toolbar-h: 48px;
            --footer-h: 28px;
        }}
        body {{
            margin: 0;
            font-family: Helvetica, Arial, sans-serif;
            background: #f4f4f4;
            color: #1e1e1e;
        }}
        header {{
            padding: 12px 16px;
            background: #1f2937;
            color: #f9fafb;
            font-size: 14px;
            letter-spacing: 0.02em;
            height: var(--header-h);
            box-sizing: border-box;
        }}
        .toolbar {{
            height: var(--toolbar-h);
            box-sizing: border-box;
            display: flex;
            align-items: center;
            gap: 8px;
            padding: 8px 12px;
            border-bottom: 1px solid #e5e7eb;
            background: #ffffff;
            overflow: hidden;
        }}
        .toolbar .spacer {{
            flex: 1;
        }}
        .toolbar input[type="search"] {{
            height: 30px;
            min-width: 280px;
            padding: 0 10px;
            border: 1px solid #d1d5db;
            border-radius: 6px;
            font-size: 12px;
        }}
        .toolbar button {{
            height: 30px;
            padding: 0 10px;
            border: 1px solid #d1d5db;
            border-radius: 6px;
            background: #ffffff;
            font-size: 12px;
            cursor: pointer;
            white-space: nowrap;
        }}
        .toolbar button:hover {{
            background: #f9fafb;
        }}
        .toolbar button.toggled {{
            border-color: #7f1d1d;
            background: #fee2e2;
            color: #7f1d1d;
        }}
        .toolbar .badge {{
            font-size: 11px;
            color: #374151;
            border: 1px solid #d1d5db;
            border-radius: 999px;
            padding: 3px 8px;
            background: #f9fafb;
            white-space: nowrap;
            max-width: 34vw;
            overflow: hidden;
            text-overflow: ellipsis;
        }}
        #inspector .node-ref {{
            border: 0;
            background: transparent;
            padding: 0;
            margin: 0;
            color: #1f2937;
            text-decoration: underline;
            cursor: pointer;
            font: inherit;
            text-align: left;
        }}
        #inspector .node-ref:hover {{
            color: #1d4ed8;
        }}
        .canvas {{
            width: 100vw;
            height: calc(100vh - var(--header-h) - var(--toolbar-h) - var(--footer-h));
            overflow: hidden;
            position: relative;
            background: radial-gradient(circle at 20% 20%, #ffffff 0%, #f4f4f4 45%, #e8e8e8 100%);
        }}
        .viewport {{
            transform-origin: 0 0;
            cursor: grab;
            transition: transform 0.03s linear;
        }}
        .viewport:active {{
            cursor: grabbing;
        }}
        #inspector {{
            position: absolute;
            top: 12px;
            right: 12px;
            width: min(360px, 40vw);
            max-height: calc(100vh - var(--header-h) - var(--toolbar-h) - 40px);
            overflow: auto;
            background: rgba(255, 255, 255, 0.95);
            border: 1px solid #d1d5db;
            border-radius: 8px;
            box-shadow: 0 6px 20px rgba(0, 0, 0, 0.12);
            padding: 10px 12px;
            font-size: 12px;
            line-height: 1.4;
            z-index: 10;
        }}
        #inspector h4 {{
            margin: 2px 0 8px;
            font-size: 12px;
        }}
        #inspector .muted {{
            color: #6b7280;
        }}
        #inspector ul {{
            margin: 4px 0 8px 16px;
            padding: 0;
        }}
        .footer {{
            height: var(--footer-h);
            box-sizing: border-box;
            border-top: 1px solid #e5e7eb;
            background: #ffffff;
            padding: 6px 12px;
            font-size: 11px;
            color: #4b5563;
            display: flex;
            align-items: center;
            gap: 12px;
        }}
        svg.focus-mode g.node,
        svg.focus-mode g.edge {{
            opacity: 0.14;
            transition: opacity 0.06s linear;
        }}
        svg.focus-mode g.node.active,
        svg.focus-mode g.edge.active {{
            opacity: 1 !important;
        }}
        g.node.search-match > polygon,
        g.node.search-match > ellipse,
        g.node.search-match > rect {{
            stroke: #059669;
            stroke-width: 2.2px;
        }}
        g.node.active > polygon,
        g.node.active > ellipse,
        g.node.active > rect {{
            stroke: #2563eb;
            stroke-width: 2.2px;
        }}
        g.edge.active > path {{
            stroke: #2563eb;
            stroke-width: 3.6px;
            opacity: 1;
        }}
        g.edge.active > polygon {{
            fill: #2563eb;
            stroke: #2563eb;
            opacity: 1;
        }}
        g.edge:not(.active) > path {{
            stroke: #D9D9D9 !important;
            stroke-width: 1.4px !important;
            stroke-opacity: 1 !important;
            opacity: 1 !important;
        }}
        g.edge:not(.active) > polygon {{
            fill: #D9D9D9 !important;
            stroke: #D9D9D9 !important;
            fill-opacity: 1 !important;
            stroke-opacity: 1 !important;
            opacity: 1 !important;
        }}
        g.edge.active {{
            opacity: 1 !important;
        }}
        g.edge.active > path {{
            stroke: #2563eb !important;
            stroke-width: 4.8px !important;
            stroke-opacity: 1 !important;
            opacity: 1 !important;
        }}
        g.edge.active > polygon {{
            fill: #2563eb !important;
            stroke: #2563eb !important;
            fill-opacity: 1 !important;
            stroke-opacity: 1 !important;
            opacity: 1 !important;
        }}
        g.cluster > path,
        g.cluster > polygon {{
            fill: none !important;
        }}
        g.cluster.cluster-dim > path,
        g.cluster.cluster-dim > polygon {{
            stroke: transparent !important;
        }}
        g.cluster.cluster-dim > text {{
            fill: transparent !important;
            stroke: transparent !important;
        }}
        g.cluster.cluster-active > path,
        g.cluster.cluster-active > polygon {{
            stroke: #000000 !important;
        }}
        svg.review-mode g.node.arch-hot > polygon,
        svg.review-mode g.node.arch-hot > ellipse,
        svg.review-mode g.node.arch-hot > rect {{
            fill: #FECACA !important;
            stroke: #B91C1C !important;
            stroke-width: 2.8px !important;
        }}
        svg.review-mode g.node.arch-warm > polygon,
        svg.review-mode g.node.arch-warm > ellipse,
        svg.review-mode g.node.arch-warm > rect {{
            fill: #FEF3C7 !important;
            stroke: #B45309 !important;
            stroke-width: 2.2px !important;
        }}
        svg.review-mode g.edge.arch-hot > path {{
            stroke: #DC2626 !important;
            stroke-width: 3.2px !important;
        }}
        svg.review-mode g.edge.arch-hot > polygon {{
            fill: #DC2626 !important;
            stroke: #DC2626 !important;
        }}
        svg.review-mode g.cluster.arch-hot > path,
        svg.review-mode g.cluster.arch-hot > polygon {{
            fill: #FECACA66 !important;
            stroke: #B91C1C !important;
            stroke-width: 2px !important;
        }}
        svg.review-mode g.cluster.arch-warm > path,
        svg.review-mode g.cluster.arch-warm > polygon {{
            fill: #FEF3C74D !important;
            stroke: #B45309 !important;
        }}
    </style>
</head>
<body>
    <header>{display_title}</header>
    <div class="toolbar">
        <button id="btn-fit" title="Fit graph to viewport (F)">Fit</button>
        <button id="btn-reset-zoom" title="Reset zoom to 100% (0)">100%</button>
        <button id="btn-zoom-in" title="Zoom in (+)">+</button>
        <button id="btn-zoom-out" title="Zoom out (-)">-</button>
        <button id="btn-chain" title="Show dependency chain (C)">Chain</button>
        <button id="btn-arch-review" title="Highlight likely bottlenecks (A)">Architecture review</button>
        <button id="btn-clear-focus" title="Clear current pin/focus (Esc)">Clear focus</button>
        <input id="search" type="search" placeholder="Search module/path… (/)">
        <span class="badge" id="search-count">0 matches</span>
        <span class="spacer"></span>
        <span class="badge" id="repo-badge">Repos: detecting…</span>
    </div>
    <div class="canvas" id="canvas">
        <div class="viewport" id="viewport">{svg}</div>
        <aside id="inspector">
            <h4>Context</h4>
            <div class="muted">Click a node or edge to pin details.</div>
        </aside>
    </div>
    <div class="footer" id="footer">Ready</div>
    <script>
        const canvas = document.getElementById('canvas');
        const viewport = document.getElementById('viewport');
        const svg = viewport.querySelector('svg');
        const inspector = document.getElementById('inspector');
        const footer = document.getElementById('footer');
        const searchInput = document.getElementById('search');
        const searchCount = document.getElementById('search-count');
        const repoBadge = document.getElementById('repo-badge');
        const initialViewBox = svg?.getAttribute('viewBox') || '';
        let scale = 1;
        let originX = 0;
        let originY = 0;
        let isDragging = false;
        let startX = 0;
        let startY = 0;
        let pinned = null;
        let activeNodeName = null;
        let searchResults = [];
        let searchIndex = -1;

        const esc = (text) => String(text)
            .replaceAll('&', '&amp;')
            .replaceAll('<', '&lt;')
            .replaceAll('>', '&gt;')
            .replaceAll('"', '&quot;')
            .replaceAll("'", '&#39;');

        const titleOf = (group) =>
            group?.querySelector('title')?.textContent?.trim() || '';
        const nodeGroups = [...svg.querySelectorAll('g.node')];
        const edgeGroups = [...svg.querySelectorAll('g.edge')];
        const clusterGroups = [...svg.querySelectorAll('g.cluster')];
        const clusterByKey = new Map();
        const nodeByName = new Map();
        const outgoing = new Map();
        const incoming = new Map();
        const edgesByNode = new Map();
        const edgesByPair = new Map();
        const nodeClusterChain = new Map();
        const nodeToDeepestClusterKey = new Map();
        const nodeClusterKeysByDepth = new Map();
        const clusterNodeNames = new Map();
        const summaryClusterMembers = new Map();
        const repos = new Set();
        const packageRoots = new Set();
        const activeNodes = new Set();
        const activeEdges = new Set();
        let hoverFrame = null;
        let inspectorHoverName = null;
        let lastFocusedNodeName = null;
        let lastClickNodeName = null;
        let lastClickAtMs = 0;
        const DOUBLE_CLICK_MS = 320;
        let reviewMode = false;
        let reviewSummary = null;

        clusterGroups.forEach((cluster) => {{
            const raw = titleOf(cluster);
            const key = raw.startsWith('cluster_') ? raw.slice(8) : raw;
            cluster.dataset.clusterKey = key;
            clusterByKey.set(key, cluster);
            clusterNodeNames.set(key, []);
        }});

        nodeGroups.forEach((node) => {{
            const name = titleOf(node);
            node.dataset.name = name;
            nodeByName.set(name, node);
            outgoing.set(name, []);
            incoming.set(name, []);
            edgesByNode.set(name, []);
            const topPackage = (name.split('.')[0] || '').trim();
            if (topPackage) packageRoots.add(topPackage);
            const path =
                node.querySelector('a')?.getAttribute('xlink:title') ||
                node.querySelector('a')?.getAttribute('title') ||
                '';
            node.dataset.path = path;
            if (path.includes('/riskstudio-worker/')) repos.add('riskstudio-worker');
            if (path.includes('/riskstudio-sdk/')) repos.add('riskstudio-sdk');

            const parts = name.split('/').filter(Boolean);
            const clusterChain = [];
            let prefix = '';
            for (const part of parts.slice(0, -1)) {{
                prefix = prefix ? `${{prefix}}/${{part}}` : part;
                const cluster = clusterByKey.get(prefix);
                if (cluster) clusterChain.push(cluster);
            }}
            nodeClusterChain.set(name, clusterChain);
            const deepest = clusterChain.length
                ? clusterChain[clusterChain.length - 1].dataset.clusterKey || ''
                : '';
            nodeToDeepestClusterKey.set(name, deepest);
            if (deepest) (clusterNodeNames.get(deepest) || []).push(name);
            const depthKeys = clusterChain.map((cluster) => cluster.dataset.clusterKey || '');
            nodeClusterKeysByDepth.set(name, depthKeys);
            depthKeys.forEach((summaryKey) => {{
                if (!summaryKey) return;
                const members = summaryClusterMembers.get(summaryKey) || [];
                members.push(name);
                summaryClusterMembers.set(summaryKey, members);
            }});
        }});

        if (repos.size) {{
            repoBadge.textContent = `Repos: ${{[...repos].sort().join(' + ')}}`;
        }} else if (packageRoots.size) {{
            repoBadge.textContent = `Packages: ${{packageRoots.size}}`;
        }} else {{
            repoBadge.textContent = 'Repos: unknown';
        }}

        edgeGroups.forEach((edge) => {{
            const raw = titleOf(edge);
            const [src, tgt] = raw.split('->').map((s) => s.trim());
            edge.dataset.src = src || '';
            edge.dataset.tgt = tgt || '';
            if (src && tgt) {{
                outgoing.get(src)?.push(tgt);
                incoming.get(tgt)?.push(src);
                edgesByNode.get(src)?.push(edge);
                edgesByNode.get(tgt)?.push(edge);
                const key = `${{src}}\u0000${{tgt}}`;
                const existing = edgesByPair.get(key) || [];
                existing.push(edge);
                edgesByPair.set(key, existing);
            }}
        }});

        const _clusterBox = (clusterKey) => {{
            const cluster = clusterByKey.get(clusterKey);
            if (!cluster) return null;
            const shape = cluster.querySelector('path, polygon');
            const box = shape ? shape.getBBox() : cluster.getBBox();
            if (box.width > 0 && box.height > 0) return box;
            const members = summaryClusterMembers.get(clusterKey) || [];
            if (!members.length) return null;
            const node = nodeByName.get(members[0]);
            if (!node) return null;
            return node.getBBox();
        }};

        const _clusterCenter = (clusterKey) => {{
            const box = _clusterBox(clusterKey);
            if (!box) return null;
            return _boxCenter(box);
        }};

        const _boxCenter = (box) => ({{
            x: box.x + box.width / 2,
            y: box.y + box.height / 2,
        }});

        const _boxBorderPoint = (box, toward) => {{
            const center = _boxCenter(box);
            let dx = toward.x - center.x;
            let dy = toward.y - center.y;
            if (Math.abs(dx) < 0.0001 && Math.abs(dy) < 0.0001) {{
                dx = 1;
                dy = 0;
            }}
            const halfW = Math.max(1, box.width / 2);
            const halfH = Math.max(1, box.height / 2);
            const tx = halfW / Math.max(0.0001, Math.abs(dx));
            const ty = halfH / Math.max(0.0001, Math.abs(dy));
            const t = Math.min(tx, ty);
            return {{
                x: center.x + dx * t,
                y: center.y + dy * t,
            }};
        }};

        const _percentileThreshold = (values, percentile, minimum) => {{
            if (!values.length) return minimum;
            const sorted = [...values].sort((a, b) => a - b);
            const index = Math.max(0, Math.min(sorted.length - 1, Math.floor(sorted.length * percentile)));
            return Math.max(minimum, sorted[index]);
        }};

        const _buildArchitectureReview = () => {{
            const nodeStats = [...nodeByName.keys()].map((name) => {{
                const inDegree = (incoming.get(name) || []).length;
                const outDegree = (outgoing.get(name) || []).length;
                const score = inDegree + outDegree + Math.sqrt(inDegree * outDegree);
                return {{ name, inDegree, outDegree, score }};
            }});
            const nodeScores = nodeStats.map((item) => item.score);
            const hotNodeThreshold = _percentileThreshold(nodeScores, 0.9, 4);
            const warmNodeThreshold = _percentileThreshold(nodeScores, 0.75, 2.5);

            nodeStats.forEach((item) => {{
                const node = nodeByName.get(item.name);
                if (!node) return;
                if (item.score >= hotNodeThreshold) {{
                    node.classList.add('arch-hot');
                }} else if (item.score >= warmNodeThreshold) {{
                    node.classList.add('arch-warm');
                }}
            }});

            const edgeStats = edgeGroups.map((edge) => {{
                const src = edge.dataset.src || '';
                const tgt = edge.dataset.tgt || '';
                const score = (outgoing.get(src) || []).length + (incoming.get(tgt) || []).length;
                return {{ edge, src, tgt, score }};
            }});
            const edgeScores = edgeStats.map((item) => item.score);
            const hotEdgeThreshold = _percentileThreshold(edgeScores, 0.9, 4);
            edgeStats.forEach((item) => {{
                if (item.score >= hotEdgeThreshold) item.edge.classList.add('arch-hot');
            }});

            const clusterScore = new Map();
            clusterGroups.forEach((cluster) => clusterScore.set(cluster, 0));
            nodeStats.forEach((item) => {{
                const chain = nodeClusterChain.get(item.name) || [];
                chain.forEach((cluster) => {{
                    clusterScore.set(cluster, (clusterScore.get(cluster) || 0) + item.score);
                }});
            }});
            const clusterValues = [...clusterScore.values()];
            const hotClusterThreshold = _percentileThreshold(clusterValues, 0.9, 6);
            const warmClusterThreshold = _percentileThreshold(clusterValues, 0.75, 3);
            clusterScore.forEach((score, cluster) => {{
                if (score >= hotClusterThreshold) {{
                    cluster.classList.add('arch-hot');
                }} else if (score >= warmClusterThreshold) {{
                    cluster.classList.add('arch-warm');
                }}
            }});

            const topNodes = [...nodeStats]
                .sort((a, b) => b.score - a.score)
                .slice(0, 6)
                .map((item) => item.name);
            const topEdges = [...edgeStats]
                .sort((a, b) => b.score - a.score)
                .slice(0, 6)
                .map((item) => `${{item.src}} -> ${{item.tgt}}`);
            const topClusters = [...clusterScore.entries()]
                .sort((a, b) => b[1] - a[1])
                .slice(0, 4)
                .map(([cluster]) => cluster.dataset.clusterKey || titleOf(cluster));
            return {{ topNodes, topEdges, topClusters, hotNodeThreshold, hotEdgeThreshold }};
        }};

        reviewSummary = _buildArchitectureReview();

        svg.querySelectorAll('a').forEach((link) => {{
            link.addEventListener('click', (event) => event.preventDefault());
        }});

        const applyTransform = () => {{
            viewport.style.transform =
                `translate(${{originX}}px, ${{originY}}px) scale(${{scale}})`;
            footer.textContent =
                `Nodes: ${{nodeGroups.length}} · ` +
                `Edges: ${{edgeGroups.length}} · ` +
                `Zoom: ${{Math.round(scale * 100)}}%`;
        }};

        const fitToView = () => {{
            if (!svg || !svg.viewBox || !svg.viewBox.baseVal) return;
            const vb = svg.viewBox.baseVal;
            if (!vb.width || !vb.height) return;
            const scaleX = canvas.clientWidth / vb.width;
            const scaleY = canvas.clientHeight / vb.height;
            scale = Math.min(scaleX, scaleY) * 0.84;
            originX = (canvas.clientWidth - vb.width * scale) / 2;
            originY = (canvas.clientHeight - vb.height * scale) / 2;
            applyTransform();
        }};

        const _syncClassSet = (previous, next, className) => {{
            previous.forEach((element) => {{
                if (!next.has(element)) element.classList.remove(className);
            }});
            next.forEach((element) => {{
                if (!previous.has(element)) element.classList.add(className);
            }});
            previous.clear();
            next.forEach((element) => previous.add(element));
        }};

        const _applyFocus = (nextNodes, nextEdges) => {{
            const hasFocus = nextNodes.size > 0 || nextEdges.size > 0;
            svg.classList.toggle('focus-mode', hasFocus);
            _syncClassSet(activeNodes, nextNodes, 'active');
            _syncClassSet(activeEdges, nextEdges, 'active');
        }};

        const _clearFocusClasses = () => {{
            svg.classList.remove('focus-mode');
            _applyFocus(new Set(), new Set());
        }};

        const _scheduleHover = (callback) => {{
            if (hoverFrame !== null) cancelAnimationFrame(hoverFrame);
            hoverFrame = requestAnimationFrame(() => {{
                hoverFrame = null;
                callback();
            }});
        }};

        const _nodeRefList = (items) =>
            `<ul>${{items.map((item) =>
                `<li><button type="button" class="node-ref" data-node-ref="${{esc(item)}}">${{esc(item)}}</button></li>`
            ).join('')}}</ul>`;

        const renderArchitectureReview = () => {{
            if (!reviewSummary) return;
            const nodeList = reviewSummary.topNodes.length
                ? _nodeRefList(reviewSummary.topNodes)
                : '<div class="muted">none</div>';
            const edgeList = reviewSummary.topEdges.length
                ? `<ul>${{reviewSummary.topEdges.map((edge) => `<li>${{esc(edge)}}</li>`).join('')}}</ul>`
                : '<div class="muted">none</div>';
            const clusterList = reviewSummary.topClusters.length
                ? `<ul>${{reviewSummary.topClusters.map((cluster) => `<li>${{esc(cluster)}}</li>`).join('')}}</ul>`
                : '<div class="muted">none</div>';
            inspector.innerHTML = `
                <h4>Architecture review</h4>
                <div class="muted">Heuristics: node centrality and edge fan-in/fan-out hotspots.</div>
                <div><strong>Hot modules</strong> (red)</div>
                ${{nodeList}}
                <div><strong>Hot dependencies</strong> (red edges)</div>
                ${{edgeList}}
                <div><strong>Hot clusters</strong> (cluster tint)</div>
                ${{clusterList}}
                <div><strong>Improvement ideas</strong></div>
                <ul>
                    <li>Split top hot modules by ownership boundary to reduce coupling.</li>
                    <li>Replace high-traffic direct dependencies with ports/adapters or events.</li>
                    <li>Move heavy integration code from core clusters into edge adapters.</li>
                </ul>
            `;
        }};

        const clearFocus = (force = false) => {{
            if (pinned && !force) return;
            _clearFocusClasses();
            clusterGroups.forEach((c) => c.classList.remove('cluster-dim', 'cluster-active'));
            activeNodeName = null;
            inspectorHoverName = null;
            if (reviewMode) {{
                renderArchitectureReview();
                return;
            }}
            inspector.innerHTML =
                '<h4>Context</h4>' +
                '<div class="muted">Click a node or edge to pin details.</div>';
        }};

        const focusNodeClusters = (name) => {{
            const activeClusters = new Set(nodeClusterChain.get(name) || []);
            if (!activeClusters.size) {{
                clusterGroups.forEach((cluster) => {{
                    cluster.classList.remove('cluster-dim', 'cluster-active');
                }});
                return;
            }}
            clusterGroups.forEach((cluster) => {{
                cluster.classList.add('cluster-dim');
                cluster.classList.remove('cluster-active');
            }});
            activeClusters.forEach((cluster) => {{
                cluster.classList.remove('cluster-dim');
                cluster.classList.add('cluster-active');
            }});
        }};

        const focusNodeClustersForNames = (names) => {{
            const activeClusters = new Set();
            names.forEach((name) => {{
                (nodeClusterChain.get(name) || []).forEach((cluster) => {{
                    activeClusters.add(cluster);
                }});
            }});
            if (!activeClusters.size) {{
                clusterGroups.forEach((cluster) => {{
                    cluster.classList.remove('cluster-dim', 'cluster-active');
                }});
                return;
            }}
            clusterGroups.forEach((cluster) => {{
                cluster.classList.add('cluster-dim');
                cluster.classList.remove('cluster-active');
            }});
            activeClusters.forEach((cluster) => {{
                cluster.classList.remove('cluster-dim');
                cluster.classList.add('cluster-active');
            }});
        }};

        const _collectNodeElements = (names) => {{
            const result = new Set();
            names.forEach((name) => {{
                const node = nodeByName.get(name);
                if (node) result.add(node);
            }});
            return result;
        }};

        const focusNode = (name, isPinned = false, updateInspector = true) => {{
            activeNodeName = name;
            lastFocusedNodeName = name;
            const out = outgoing.get(name) || [];
            const inc = incoming.get(name) || [];
            const neighbors = new Set([name, ...out, ...inc]);
            const nextNodes = _collectNodeElements(neighbors);
            const nextEdges = new Set(edgesByNode.get(name) || []);
            _applyFocus(nextNodes, nextEdges);
            focusNodeClusters(name);

            if (!updateInspector) return;

            const outList = out.length
                ? _nodeRefList(out)
                : '<div class="muted">none</div>';
            const inList = inc.length
                ? _nodeRefList(inc)
                : '<div class="muted">none</div>';
            const pinBadge = isPinned
                ? '<div class="muted"><strong>Pinned</strong> · click empty space or press Esc to release</div>'
                : '';

            inspector.innerHTML = `
                <h4>${{esc(name)}}</h4>
                ${{pinBadge}}
                <div><strong>Outgoing</strong> (${{out.length}})</div>
                ${{outList}}
                <div><strong>Incoming</strong> (${{inc.length}})</div>
                ${{inList}}
            `;
        }};

        const _collectTransitive = (seed, adjacency) => {{
            const visited = new Set([seed]);
            const queue = [seed];
            while (queue.length) {{
                const node = queue.shift();
                const next = adjacency.get(node) || [];
                next.forEach((candidate) => {{
                    if (visited.has(candidate)) return;
                    visited.add(candidate);
                    queue.push(candidate);
                }});
            }}
            return visited;
        }};

        const focusNodeTransitive = (name, isPinned = false, updateInspector = true) => {{
            activeNodeName = name;
            lastFocusedNodeName = name;
            const descendants = _collectTransitive(name, outgoing);
            const ancestors = _collectTransitive(name, incoming);
            const activeNames = new Set([...descendants, ...ancestors]);
            const nextNodes = _collectNodeElements(activeNames);
            const nextEdges = new Set();
            edgeGroups.forEach((edge) => {{
                const src = edge.dataset.src;
                const tgt = edge.dataset.tgt;
                if (activeNames.has(src) && activeNames.has(tgt)) {{
                    nextEdges.add(edge);
                }}
            }});
            _applyFocus(nextNodes, nextEdges);
            focusNodeClustersForNames(activeNames);

            if (!updateInspector) return;
            const pinBadge = isPinned
                ? '<div class="muted"><strong>Pinned</strong> · transitief (parents + children, recursief)</div>'
                : '';
            inspector.innerHTML = `
                <h4>${{esc(name)}}</h4>
                ${{pinBadge}}
                <div><strong>Transitive context</strong></div>
                <div>Nodes: ${{activeNames.size}}</div>
                <div>Ancestors: ${{Math.max(0, ancestors.size - 1)}}</div>
                <div>Descendants: ${{Math.max(0, descendants.size - 1)}}</div>
            `;
        }};

        const restorePinnedView = () => {{
            if (!pinned) {{
                clearFocus(true);
                return;
            }}
            if (pinned.kind === 'node') {{
                focusNode(pinned.name, true);
                return;
            }}
            if (pinned.kind === 'node-transitive') {{
                focusNodeTransitive(pinned.name, true);
                return;
            }}
            focusEdge(pinned.src, pinned.tgt, true);
        }};

        const focusEdge = (src, tgt, isPinned = false) => {{
            activeNodeName = null;
            const nextNodes = _collectNodeElements(new Set([src, tgt]));
            const key = `${{src}}\u0000${{tgt}}`;
            const nextEdges = new Set(edgesByPair.get(key) || []);
            _applyFocus(nextNodes, nextEdges);
            clusterGroups.forEach((c) => c.classList.remove('cluster-dim', 'cluster-active'));

            const pinBadge = isPinned
                ? '<div class="muted"><strong>Pinned</strong> · click empty space or press Esc to release</div>'
                : '';

            inspector.innerHTML = `
                <h4>Dependency</h4>
                ${{pinBadge}}
                <div><strong>From:</strong> ${{esc(src)}}</div>
                <div><strong>To:</strong> ${{esc(tgt)}}</div>
            `;
        }};

        const triggerDependencyChain = () => {{
            const chainName = pinned?.kind === 'node' || pinned?.kind === 'node-transitive'
                ? pinned.name
                : (activeNodeName || lastFocusedNodeName);
            if (!chainName) return;
            pinned = {{ kind: 'node-transitive', name: chainName }};
            focusNodeTransitive(chainName, true);
        }};

        const setArchitectureReviewMode = (enabled) => {{
            reviewMode = enabled;
            svg.classList.toggle('review-mode', reviewMode);
            const reviewButton = document.getElementById('btn-arch-review');
            reviewButton.classList.toggle('toggled', reviewMode);
            if (reviewMode) {{
                renderArchitectureReview();
                return;
            }}
            restorePinnedView();
        }};

        const clearSearchStyles = () => {{
            nodeGroups.forEach((n) => n.classList.remove('search-match'));
        }};

        const runSearch = (query) => {{
            const q = query.trim().toLowerCase();
            clearSearchStyles();
            searchResults = [];
            searchIndex = -1;
            if (!q) {{
                searchCount.textContent = '0 matches';
                return;
            }}

            for (const node of nodeGroups) {{
                const name = (node.dataset.name || '').toLowerCase();
                const path = (node.dataset.path || '').toLowerCase();
                if (name.includes(q) || path.includes(q)) {{
                    node.classList.add('search-match');
                    searchResults.push(node);
                }}
            }}

            searchCount.textContent =
                `${{searchResults.length}} match${{searchResults.length === 1 ? '' : 'es'}}`;
            if (searchResults.length > 0) {{
                searchIndex = 0;
                const name = searchResults[0].dataset.name;
                focusNode(name, false);
            }}
        }};

        const stepSearch = () => {{
            if (!searchResults.length) return;
            searchIndex = (searchIndex + 1) % searchResults.length;
            const node = searchResults[searchIndex];
            const name = node.dataset.name;
            focusNode(name, false);
        }};

        nodeGroups.forEach((node) => {{
            const name = node.dataset.name;
            node.addEventListener('mouseenter', () => {{
                if (pinned) return;
                _scheduleHover(() => focusNode(name));
            }});
            node.addEventListener('mouseleave', () => _scheduleHover(() => clearFocus()));
            node.addEventListener('click', (event) => {{
                event.preventDefault();
                event.stopPropagation();
                const nowMs = Date.now();
                const isRapidSecondClick =
                    lastClickNodeName === name && (nowMs - lastClickAtMs) <= DOUBLE_CLICK_MS;
                lastClickNodeName = name;
                lastClickAtMs = nowMs;
                if (isRapidSecondClick) {{
                    pinned = {{ kind: 'node-transitive', name }};
                    focusNodeTransitive(name, true);
                    return;
                }}
                pinned = {{ kind: 'node', name }};
                focusNode(name, true);
            }});
            node.addEventListener('dblclick', (event) => {{
                event.preventDefault();
                event.stopPropagation();
                pinned = {{ kind: 'node-transitive', name }};
                focusNodeTransitive(name, true);
            }});
        }});

        edgeGroups.forEach((edge) => {{
            const src = edge.dataset.src;
            const tgt = edge.dataset.tgt;
            if (!src || !tgt) return;
            edge.addEventListener('mouseenter', () => {{
                if (pinned) return;
                _scheduleHover(() => focusEdge(src, tgt));
            }});
            edge.addEventListener('mouseleave', () => _scheduleHover(() => clearFocus()));
            edge.addEventListener('click', (event) => {{
                event.preventDefault();
                event.stopPropagation();
                pinned = {{ kind: 'edge', src, tgt }};
                focusEdge(src, tgt, true);
            }});
        }});

        inspector.addEventListener('click', (event) => {{
            const target = event.target instanceof Element ? event.target : null;
            const nodeRef = target ? target.closest('[data-node-ref]') : null;
            if (!nodeRef) return;
            event.preventDefault();
            event.stopPropagation();
            const name = nodeRef.getAttribute('data-node-ref') || '';
            if (!name) return;
            pinned = {{ kind: 'node', name }};
            focusNode(name, true);
        }});

        inspector.addEventListener('mouseover', (event) => {{
            const target = event.target instanceof Element ? event.target : null;
            const nodeRef = target ? target.closest('[data-node-ref]') : null;
            if (!nodeRef) return;
            const name = nodeRef.getAttribute('data-node-ref') || '';
            if (!name) return;
            if (name === inspectorHoverName) return;
            inspectorHoverName = name;
            _scheduleHover(() => focusNode(name, false, false));
        }});

        inspector.addEventListener('mouseout', (event) => {{
            const target = event.target instanceof Element ? event.target : null;
            const nodeRef = target ? target.closest('[data-node-ref]') : null;
            if (!nodeRef) return;
            const related = event.relatedTarget;
            if (related instanceof Node && nodeRef.contains(related)) return;
            inspectorHoverName = null;
            _scheduleHover(() => restorePinnedView());
        }});

        canvas.addEventListener('click', (event) => {{
            if (event.target.closest('g.node') || event.target.closest('g.edge')) return;
            pinned = null;
            clearFocus(true);
        }});

        window.addEventListener('keydown', (event) => {{
            if (event.key === 'Escape') {{
                pinned = null;
                clearFocus(true);
                return;
            }}
            if (event.key === '/' && document.activeElement !== searchInput) {{
                event.preventDefault();
                searchInput.focus();
                searchInput.select();
                return;
            }}
            const typingTarget = event.target;
            const isTyping =
                typingTarget instanceof HTMLInputElement ||
                typingTarget instanceof HTMLTextAreaElement ||
                (typingTarget instanceof HTMLElement && typingTarget.isContentEditable);
            if (!isTyping && (event.key === 'a' || event.key === 'A')) {{
                event.preventDefault();
                setArchitectureReviewMode(!reviewMode);
                return;
            }}
            if (!isTyping && (event.key === 'c' || event.key === 'C')) {{
                event.preventDefault();
                triggerDependencyChain();
                return;
            }}
            if (event.key === 'f' || event.key === 'F') {{
                fitToView();
                return;
            }}
            if (event.key === '+' || event.key === '=') {{
                scale = Math.min(4, scale + 0.1);
                applyTransform();
                return;
            }}
            if (event.key === '-') {{
                scale = Math.max(0.2, scale - 0.1);
                applyTransform();
                return;
            }}
            if (event.key === '0') {{
                scale = 1;
                applyTransform();
                return;
            }}
        }});

        canvas.addEventListener('wheel', (event) => {{
            event.preventDefault();
            const delta = Math.sign(event.deltaY) * -0.1;
            scale = Math.min(4, Math.max(0.2, scale + delta));
            applyTransform();
        }});

        canvas.addEventListener('mousedown', (event) => {{
            isDragging = true;
            startX = event.clientX - originX;
            startY = event.clientY - originY;
        }});

        window.addEventListener('mousemove', (event) => {{
            if (!isDragging) return;
            originX = event.clientX - startX;
            originY = event.clientY - startY;
            applyTransform();
        }});

        window.addEventListener('mouseup', () => {{
            isDragging = false;
        }});

        searchInput.addEventListener('input', () => runSearch(searchInput.value));
        searchInput.addEventListener('keydown', (event) => {{
            if (event.key === 'Enter') {{
                event.preventDefault();
                stepSearch();
            }}
        }});

        document.getElementById('btn-fit').addEventListener('click', fitToView);
        document.getElementById('btn-reset-zoom').addEventListener('click', () => {{
            scale = 1;
            applyTransform();
        }});
        document.getElementById('btn-zoom-in').addEventListener('click', () => {{
            scale = Math.min(4, scale + 0.1);
            applyTransform();
        }});
        document.getElementById('btn-zoom-out').addEventListener('click', () => {{
            scale = Math.max(0.2, scale - 0.1);
            applyTransform();
        }});
        document.getElementById('btn-chain').addEventListener('click', () => {{
            triggerDependencyChain();
        }});
        document.getElementById('btn-arch-review').addEventListener('click', () => {{
            setArchitectureReviewMode(!reviewMode);
        }});
        document.getElementById('btn-clear-focus').addEventListener('click', () => {{
            pinned = null;
            clearFocus(true);
        }});

        window.addEventListener('resize', fitToView);
        fitToView();
    </script>
</body>
</html>"""


def _html_with_fallback(dot: str, title: str, error: str) -> str:
    display_title = _display_graph_title(title)
    return f"""<!doctype html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1">
    <title>{display_title}</title>
  <style>
    body {{
      margin: 0;
      font-family: Helvetica, Arial, sans-serif;
      background: #f4f4f4;
      color: #1e1e1e;
    }}
    header {{
      padding: 12px 16px;
      background: #1f2937;
      color: #f9fafb;
      font-size: 14px;
    }}
    .warning {{
      padding: 12px 16px;
      background: #fff3cd;
      color: #5c4400;
      border-bottom: 1px solid #e8d18a;
    }}
    pre {{
      padding: 16px;
      white-space: pre-wrap;
    }}
  </style>
</head>
<body>
    <header>{display_title}</header>
  <div class="warning">Graphviz rendering failed: {error}</div>
  <pre>{dot}</pre>
</body>
</html>"""


def _display_graph_title(title: str) -> str:
    return title.replace("import_cruiser", "import-cruiser")



"""Detect circular dependencies using DFS."""

from __future__ import annotations

from import_cruiser.graph import DependencyGraph, detect_cycles as _detect_cycles


def detect_cycles(graph: DependencyGraph) -> list[list[str]]:
    """Backward-compatible wrapper around graph.detect_cycles."""
    cycles = _detect_cycles(graph)
    return [list(cycle) for cycle in cycles]



"""Validate a DependencyGraph against user-defined rules."""

from __future__ import annotations

import re

from import_cruiser.config import JSONDict
from import_cruiser.graph import DependencyGraph


def _matches_pattern(name: str, pattern_obj: JSONDict) -> bool:
    """Return True if *name* matches the pattern specification in *pattern_obj*.

    Supported keys:
      - ``path``: a regex applied to the module name (dot-separated)
    """
    path_pattern = pattern_obj.get("path")
    if isinstance(path_pattern, str):
        return bool(re.search(path_pattern, name))
    # Empty pattern matches everything
    return True


class Violation:
    """A single rule violation."""

    __slots__ = ("rule_name", "severity", "message", "source", "target")

    def __init__(
        self,
        rule_name: str,
        severity: str,
        message: str,
        source: str,
        target: str,
    ) -> None:
        self.rule_name = rule_name
        self.severity = severity
        self.message = message
        self.source = source
        self.target = target

    def to_dict(self) -> JSONDict:
        return {
            "rule": self.rule_name,
            "severity": self.severity,
            "message": self.message,
            "source": self.source,
            "target": self.target,
        }


class Validator:
    """Validate a DependencyGraph against a list of rules."""

    def __init__(self, rules: list[JSONDict]) -> None:
        self.rules = rules

    def validate(self, graph: DependencyGraph) -> list[Violation]:
        """Run all rules against *graph* and return a list of violations."""
        violations: list[Violation] = []
        for rule in self.rules:
            violations.extend(self._apply_rule(rule, graph))
        return violations

    def _apply_rule(self, rule: JSONDict, graph: DependencyGraph) -> list[Violation]:
        violations: list[Violation] = []
        rule_name = str(rule.get("name", "rule"))
        severity = str(rule.get("severity", "error"))
        from_raw = rule.get("from", {})
        to_raw = rule.get("to", {})
        from_pattern = from_raw if isinstance(from_raw, dict) else {}
        to_pattern = to_raw if isinstance(to_raw, dict) else {}
        allow = bool(rule.get("allow", True))

        for dep in graph.dependencies:
            source_matches = _matches_pattern(dep.source, from_pattern)
            target_matches = _matches_pattern(dep.target, to_pattern)

            if source_matches and target_matches:
                if not allow:
                    msg = (
                        f"Dependency from '{dep.source}' to '{dep.target}' "
                        f"is forbidden by rule '{rule_name}'."
                    )
                    violations.append(
                        Violation(
                            rule_name=rule_name,
                            severity=severity,
                            message=msg,
                            source=dep.source,
                            target=dep.target,
                        )
                    )

        return violations



"""Parse Python source files and build a DependencyGraph."""

from __future__ import annotations

import ast
import os
import re
from urllib.parse import urlsplit
import warnings
from pathlib import Path

from import_cruiser.graph import Dependency, DependencyGraph, Module


def _module_name_from_path(
    path: Path, base: Path, normalize_hyphens: bool = True
) -> str:
    """Convert a file path to a dotted module name relative to *base*."""
    rel = path.relative_to(base)
    parts = list(rel.parts)
    if normalize_hyphens:
        parts = [p.replace("-", "_") for p in parts]
    if parts[-1] == "__init__.py":
        parts = parts[:-1]
    else:
        parts[-1] = parts[-1].removesuffix(".py")
    return ".".join(parts)


def _collect_imports(source: str, module_name: str) -> list[tuple[str, int]]:
    """Return list of (import_name, line_no) extracted from *source*."""
    try:
        # Some real-world files trigger SyntaxWarning (e.g. invalid escape
        # sequences in strings); keep CLI output quiet by default.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", SyntaxWarning)
            tree = ast.parse(source)
    except SyntaxError:
        return []

    imports: list[tuple[str, int]] = []
    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                imports.append((alias.name, node.lineno))
        elif isinstance(node, ast.ImportFrom):
            if node.level == 0:
                # Absolute import: `from pkg import name`
                if node.module:
                    imports.append((node.module, node.lineno))
                else:
                    # `from . import name` with level 0 shouldn't happen, skip
                    pass
            else:
                # Relative import – resolve against current module
                parts = module_name.split(".")
                base_parts = parts[: max(0, len(parts) - node.level)]
                if node.module:
                    resolved = (
                        ".".join(base_parts + [node.module])
                        if base_parts
                        else node.module
                    )
                    imports.append((resolved, node.lineno))
                else:
                    # `from . import name` – each name is a sub-module
                    for alias in node.names:
                        resolved = (
                            ".".join(base_parts + [alias.name])
                            if base_parts
                            else alias.name
                        )
                        imports.append((resolved, node.lineno))

    if imports:
        return imports

    parts = module_name.split(".")
    for lineno, line in enumerate(source.splitlines(), start=1):
        stripped = line.strip()
        if not stripped.startswith("from ") or " import " not in stripped:
            continue
        target = stripped[5:].split(" import ", 1)[0].strip()
        if not target.startswith("."):
            continue

        level = len(target) - len(target.lstrip("."))
        module = target[level:]
        base_parts = parts[: max(0, len(parts) - level)]
        resolved = ".".join(base_parts + [module]) if base_parts and module else module
        if resolved:
            imports.append((resolved, lineno))

    return imports


def _iter_python_files(
    directory: Path,
    include_paths: list[re.Pattern[str]] | None = None,
    exclude_paths: list[re.Pattern[str]] | None = None,
) -> list[Path]:
    py_files: list[Path] = []
    for dirpath, dirnames, filenames in os.walk(directory):
        # Skip hidden directories and common noise
        dirnames[:] = [
            d for d in dirnames if not d.startswith(".") and d not in ("__pycache__",)
        ]
        for filename in filenames:
            if filename.endswith(".py"):
                path = Path(dirpath) / filename
                path_text = str(path).replace("\\", "/")
                if include_paths and not any(
                    pattern.search(path_text) for pattern in include_paths
                ):
                    continue
                if exclude_paths and any(
                    pattern.search(path_text) for pattern in exclude_paths
                ):
                    continue
                py_files.append(path)
    return py_files


class Analyzer:
    """Analyze a Python project directory and produce a DependencyGraph."""

    def __init__(
        self,
        root: str | Path,
        normalize_hyphens: bool = True,
        include_paths: list[str] | None = None,
        exclude_paths: list[str] | None = None,
        include_external_patterns: list[str] | None = None,
        include_http_hosts: bool = False,
    ) -> None:
        # Keep the user-provided path form (e.g. /private/tmp vs /tmp) so
        # include/exclude path regexes match what users pass on the CLI.
        self.root = Path(root).expanduser().absolute()
        self.normalize_hyphens = normalize_hyphens
        self.include_paths = [re.compile(p) for p in include_paths or []]
        self.exclude_paths = [re.compile(p) for p in exclude_paths or []]
        self.include_external_patterns = [
            re.compile(p) for p in include_external_patterns or []
        ]
        self.include_http_hosts = include_http_hosts

    def analyze(self) -> DependencyGraph:
        """Walk the project directory and return a populated graph."""
        graph = DependencyGraph()
        py_files = _iter_python_files(
            self.root,
            include_paths=self.include_paths or None,
            exclude_paths=self.exclude_paths or None,
        )
        source_roots = _find_source_roots(self.root)

        # First pass: register all modules
        module_map: dict[str, Path] = {}
        for py_file in py_files:
            base = _select_source_root(py_file, source_roots) or self.root
            mod_name = _module_name_from_path(
                py_file,
                base,
                self.normalize_hyphens,
            )
            module_map[mod_name] = py_file
            graph.add_module(Module(name=mod_name, path=str(py_file)))

        known_modules = set(module_map.keys())

        # Second pass: resolve imports → edges
        for mod_name, py_file in module_map.items():
            source = py_file.read_text(encoding="utf-8", errors="replace")
            raw_imports = _collect_imports(source, mod_name)
            for imported, lineno in raw_imports:
                # Only keep dependencies that are internal to the project
                target = _resolve_internal(imported, known_modules)
                if target is not None:
                    graph.add_dependency(
                        Dependency(source=mod_name, target=target, line=lineno)
                    )
                    continue

                external = _resolve_external(
                    imported,
                    self.include_external_patterns,
                )
                if external is None:
                    continue
                if graph.get_module(external) is None:
                    graph.add_module(Module(name=external, path=""))
                graph.add_dependency(
                    Dependency(source=mod_name, target=external, line=lineno)
                )

            if not self.include_http_hosts:
                continue
            for host, lineno in _collect_http_hosts(source):
                if graph.get_module(host) is None:
                    graph.add_module(Module(name=host, path=""))
                graph.add_dependency(
                    Dependency(source=mod_name, target=host, line=lineno)
                )

        return graph


def _resolve_internal(imported: str, known_modules: set[str]) -> str | None:
    """Return the best-matching known module for *imported*, or None."""
    if imported in known_modules:
        return imported

    if any(mod.startswith(imported + ".") for mod in known_modules):
        return imported

    parts = imported.split(".")
    for i in range(len(parts) - 1, 0, -1):
        candidate = ".".join(parts[:i])
        if candidate in known_modules:
            return candidate
    return None


def _resolve_external(
    imported: str,
    patterns: list[re.Pattern[str]],
) -> str | None:
    if not patterns:
        return None
    if not any(pattern.search(imported) for pattern in patterns):
        return None
    return imported.split(".", 1)[0]


def _collect_http_hosts(source: str) -> list[tuple[str, int]]:
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", SyntaxWarning)
            tree = ast.parse(source)
    except SyntaxError:
        return []

    collector = _HttpHostCollector()
    collector.visit(tree)
    return collector.hosts


class _HttpHostCollector(ast.NodeVisitor):
    def __init__(self) -> None:
        self.hosts: list[tuple[str, int]] = []
        self._scopes: list[dict[str, str]] = [{}]

    def visit_Assign(self, node: ast.Assign) -> None:
        host = _http_host_from_expr(node.value, self._scopes)
        if host:
            for target in node.targets:
                self._bind_target(target, host)
                if _is_url_like_target(target):
                    self.hosts.append((host, getattr(node, "lineno", 0)))
        self.generic_visit(node)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        if node.value is not None:
            host = _http_host_from_expr(node.value, self._scopes)
            if host:
                self._bind_target(node.target, host)
                if _is_url_like_target(node.target):
                    self.hosts.append((host, getattr(node, "lineno", 0)))
        self.generic_visit(node)

    def visit_Call(self, node: ast.Call) -> None:
        values = list(node.args)
        values.extend(kw.value for kw in node.keywords if kw.arg is not None)
        for value in values:
            host = _http_host_from_expr(value, self._scopes)
            if host:
                lineno = getattr(value, "lineno", node.lineno)
                self.hosts.append((host, lineno))
        self.generic_visit(node)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self._visit_scoped(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self._visit_scoped(node)

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        self._visit_scoped(node)

    def _visit_scoped(
        self,
        node: ast.FunctionDef | ast.AsyncFunctionDef | ast.ClassDef,
    ) -> None:
        self._scopes.append({})
        for child in node.body:
            self.visit(child)
        self._scopes.pop()

    def _bind_target(self, target: ast.expr, host: str) -> None:
        if isinstance(target, ast.Name):
            self._scopes[-1][target.id] = host


def _http_host_from_text(value: str) -> str | None:
    text = value.strip()
    if not (text.startswith("http://") or text.startswith("https://")):
        return None
    parsed = urlsplit(text)
    if not parsed.netloc:
        return None
    host = parsed.netloc.split("@")[-1].split(":")[0].lower().strip()
    if not host:
        return None
    return host


def _http_host_from_expr(expr: ast.AST, scopes: list[dict[str, str]]) -> str | None:
    if isinstance(expr, ast.Constant) and isinstance(expr.value, str):
        return _http_host_from_text(expr.value)

    if isinstance(expr, ast.JoinedStr):
        rendered = _render_joined_str(expr, scopes)
        if rendered is None:
            return None
        return _http_host_from_text(rendered)

    if isinstance(expr, ast.Name):
        return _lookup_scoped_host(expr.id, scopes)

    return None


def _render_joined_str(expr: ast.JoinedStr, scopes: list[dict[str, str]]) -> str | None:
    pieces: list[str] = []
    for value in expr.values:
        if isinstance(value, ast.Constant) and isinstance(value.value, str):
            pieces.append(value.value)
            continue
        if isinstance(value, ast.FormattedValue):
            if isinstance(value.value, ast.Name):
                resolved = _lookup_scoped_host(value.value.id, scopes)
                if resolved:
                    current = "".join(pieces)
                    if current.endswith("http://") or current.endswith("https://"):
                        pieces.append(resolved)
                    elif "http://" in current or "https://" in current:
                        pieces.append(resolved)
                    else:
                        pieces.append(f"https://{resolved}")
                    continue
            pieces.append("x")
            continue
        return None
    return "".join(pieces)


def _lookup_scoped_host(name: str, scopes: list[dict[str, str]]) -> str | None:
    for scope in reversed(scopes):
        host = scope.get(name)
        if host:
            return host
    return None


def _is_url_like_target(target: ast.expr) -> bool:
    if not isinstance(target, ast.Name):
        return False
    lowered = target.id.lower()
    return "url" in lowered or "uri" in lowered or "endpoint" in lowered


def _find_source_roots(root: Path) -> list[Path]:
    roots: list[Path] = []
    for dirpath, dirnames, filenames in os.walk(root):
        if os.path.basename(dirpath) == "src":
            src_path = Path(dirpath).absolute()
            if _contains_python_files(src_path):
                roots.append(src_path)
    return roots


def _contains_python_files(directory: Path) -> bool:
    for dirpath, _, filenames in os.walk(directory):
        if any(name.endswith(".py") for name in filenames):
            return True
    return False


def _select_source_root(path: Path, roots: list[Path]) -> Path | None:
    candidates = [r for r in roots if path.is_relative_to(r)]
    if not candidates:
        return None
    return max(candidates, key=lambda p: len(p.parts))



"""Load and validate import_cruiser configuration."""

from __future__ import annotations

import json
from pathlib import Path
from typing import TypeAlias


JSONScalar: TypeAlias = str | int | float | bool | None
JSONValue: TypeAlias = JSONScalar | list["JSONValue"] | dict[str, "JSONValue"]
JSONDict: TypeAlias = dict[str, JSONValue]

DEFAULT_CONFIG: JSONDict = {
    "rules": [],
    "options": {
        "include_external": False,
    },
}

# JSON Schema (subset) for a single rule
RULE_REQUIRED_FIELDS = {"name", "severity", "from", "to"}
VALID_SEVERITIES = {"error", "warn", "info"}


class ConfigError(ValueError):
    """Raised when the configuration is invalid."""


def load_config(path: str | Path) -> JSONDict:
    """Load a JSON configuration file and return the parsed dict."""
    config_path = Path(path)
    if not config_path.exists():
        raise ConfigError(f"Configuration file not found: {config_path}")
    try:
        with config_path.open(encoding="utf-8") as fh:
            raw = json.load(fh)
    except json.JSONDecodeError as exc:
        raise ConfigError(f"Invalid JSON in configuration file: {exc}") from exc

    validate_config(raw)
    return raw


def validate_config(config: JSONDict) -> None:
    """Raise ConfigError if *config* does not conform to the expected schema."""
    if not isinstance(config, dict):
        raise ConfigError("Configuration must be a JSON object.")

    rules = config.get("rules", [])
    if not isinstance(rules, list):
        raise ConfigError("'rules' must be a list.")

    for i, rule in enumerate(rules):
        if not isinstance(rule, dict):
            raise ConfigError(f"Rule #{i} must be an object.")
        missing = RULE_REQUIRED_FIELDS - rule.keys()
        if missing:
            raise ConfigError(f"Rule #{i} is missing required fields: {missing}")
        if rule["severity"] not in VALID_SEVERITIES:
            raise ConfigError(
                f"Rule #{i} has invalid severity '{rule['severity']}'. "
                f"Must be one of {VALID_SEVERITIES}."
            )
        for field in ("from", "to"):
            if not isinstance(rule[field], dict):
                raise ConfigError(f"Rule #{i} '{field}' must be an object.")


def default_config() -> JSONDict:
    """Return a deep copy of the default configuration."""
    import copy

    return copy.deepcopy(DEFAULT_CONFIG)



"""CLI entry point for import-cruiser."""
# pylint: disable=duplicate-code

from __future__ import annotations

import sys
import re
from pathlib import Path
from typing import Optional, cast

import click

from import_cruiser import __version__
from import_cruiser.analyzer import Analyzer
from import_cruiser.config import JSONDict, ConfigError, default_config, load_config
from import_cruiser.exporter import (
    ViolationLike,
    export_dot,
    export_html,
    export_json,
    export_svg,
)
from import_cruiser.graph import (
    DependencyGraph,
    aggregate_by_path,
    collapse_graph,
    detect_cycles,
    filter_graph,
    prune_isolated_modules,
    prune_orphan_init_modules,
)
from import_cruiser.validator import Validator, Violation

DEFAULT_NOISE_PATH_PATTERNS: tuple[str, ...] = (
    r"__init__\.py$",
    r"/\.venv/",
    r"/site-packages/",
    r"/tests/",
    r"/scripts/",
    r"/examples/",
    r"/stress/",
    r"/migrations/",
)

DB_CONNECTOR_IMPORT_PATTERNS: tuple[str, ...] = (
    r"(^|\.)sqlalchemy(\.|$)",
    r"(^|\.)sqlmodel(\.|$)",
    r"(^|\.)alembic(\.|$)",
    r"(^|\.)psycopg(2)?(\.|$)",
    r"(^|\.)asyncpg(\.|$)",
    r"(^|\.)pg8000(\.|$)",
    r"(^|\.)aiopg(\.|$)",
    r"(^|\.)databases(\.|$)",
    r"(^|\.)postgres(\.|$)",
)


@click.group()
@click.version_option(__version__, prog_name="import-cruiser")
def main() -> None:
    """import-cruiser – Analyze, validate, and visualize Python import dependencies."""


# ---------------------------------------------------------------------------
# analyze
# ---------------------------------------------------------------------------


@main.command("analyze")
@click.argument("path", default=".", type=click.Path(exists=True, file_okay=False))
@click.option(
    "--output",
    "-o",
    type=click.Path(dir_okay=False, writable=True),
    default=None,
    help="Write output to a file instead of stdout.",
)
@click.option(
    "--format",
    "-f",
    "fmt",
    type=click.Choice(["json", "dot", "svg", "html"], case_sensitive=False),
    default="json",
    show_default=True,
    help="Output format.",
)
@click.option(
    "--normalize-hyphens/--no-normalize-hyphens",
    default=True,
    show_default=True,
    help="Normalize '-' to '_' in module names.",
)
@click.option(
    "--layout",
    type=click.Choice(["dot", "sfdp", "neato", "fdp"], case_sensitive=False),
    default="dot",
    show_default=True,
    help="Graphviz layout engine for svg/html output.",
)
@click.option(
    "--style",
    type=click.Choice(
        ["default", "archi", "cruiser", "navigator", "depcruise"],
        case_sensitive=False,
    ),
    default="depcruise",
    show_default=True,
    help="Graph styling preset.",
)
@click.option(
    "--rankdir",
    type=click.Choice(["LR", "TB"], case_sensitive=False),
    default="LR",
    show_default=True,
    help="Graph direction for dot layouts.",
)
@click.option(
    "--include",
    multiple=True,
    help="Regex to include module names (repeatable).",
)
@click.option(
    "--exclude",
    multiple=True,
    help="Regex to exclude module names (repeatable).",
)
@click.option(
    "--include-path",
    multiple=True,
    help="Regex to include file paths (repeatable).",
)
@click.option(
    "--exclude-path",
    multiple=True,
    help="Regex to exclude file paths (repeatable).",
)
@click.option(
    "--exclude-common-noise-paths/--no-exclude-common-noise-paths",
    default=False,
    show_default=True,
    help=(
        "Exclude common noise paths "
        "(tests/scripts/examples/stress/migrations) from graph generation."
    ),
)
@click.option(
    "--focus",
    multiple=True,
    help="Regex to focus on module names and neighbors (repeatable).",
)
@click.option(
    "--focus-depth",
    type=int,
    default=1,
    show_default=True,
    help="Neighbor depth for focus selection.",
)
@click.option(
    "--collapse-depth",
    type=int,
    default=0,
    show_default=True,
    help="Collapse modules to a package depth (0 disables).",
)
@click.option(
    "--cluster-depth",
    type=int,
    default=5,
    show_default=True,
    help="Cluster modules by package depth (0 disables).",
)
@click.option(
    "--cluster-mode",
    type=click.Choice(["path", "module"], case_sensitive=False),
    default="path",
    show_default=True,
    help="Cluster by filesystem path or module name.",
)
@click.option(
    "--aggregate-depth",
    type=int,
    default=8,
    show_default=True,
    help="Aggregate modules by path depth (0 disables).",
)
@click.option(
    "--leaf-pattern",
    multiple=True,
    help="Regex for leaf filenames to keep as nodes (repeatable).",
)
@click.option(
    "--edge-mode",
    type=click.Choice(["node", "cluster", "auto"], case_sensitive=False),
    default="auto",
    show_default=True,
    help="Edge rendering mode for graphs.",
)
@click.option(
    "--prune-isolated/--keep-isolated",
    default=False,
    show_default=True,
    help="Drop modules without incoming/outgoing edges after filtering.",
)
@click.option(
    "--include-db-connectors/--no-include-db-connectors",
    default=False,
    show_default=True,
    help="Include DB connector imports as external nodes.",
)
@click.option(
    "--include-http-hosts/--no-include-http-hosts",
    default=False,
    show_default=True,
    help="Include HTTP request hosts as external nodes.",
)
def cmd_analyze(
    path: str,
    output: Optional[str],
    fmt: str,
    normalize_hyphens: bool,
    layout: str,
    style: str,
    rankdir: str,
    include: tuple[str, ...],
    exclude: tuple[str, ...],
    include_path: tuple[str, ...],
    exclude_path: tuple[str, ...],
    exclude_common_noise_paths: bool,
    focus: tuple[str, ...],
    focus_depth: int,
    collapse_depth: int,
    cluster_depth: int,
    cluster_mode: str,
    aggregate_depth: int,
    leaf_pattern: tuple[str, ...],
    edge_mode: str,
    prune_isolated: bool,
    include_db_connectors: bool,
    include_http_hosts: bool,
) -> None:
    """Analyze Python import dependencies in PATH and output results.

    PATH defaults to the current working directory.
    """
    effective_exclude_paths = _effective_exclude_paths(
        exclude_path,
        exclude_common_noise_paths,
    )
    graph = Analyzer(
        path,
        normalize_hyphens=normalize_hyphens,
        include_paths=list(include_path),
        exclude_paths=effective_exclude_paths,
        include_external_patterns=_external_patterns_for_db(include_db_connectors),
        include_http_hosts=include_http_hosts,
    ).analyze()
    graph, layout, rankdir, cluster_depth, cluster_mode, style, edge_mode = (
        _apply_graph_options(
            graph,
            include=list(include),
            exclude=list(exclude),
            include_paths=list(include_path),
            exclude_paths=effective_exclude_paths,
            focus=list(focus),
            focus_depth=focus_depth,
            collapse_depth=collapse_depth,
            cluster_depth=cluster_depth,
            cluster_mode=cluster_mode,
            aggregate_depth=aggregate_depth,
            leaf_patterns=list(leaf_pattern),
            layout=layout,
            rankdir=rankdir,
            style=style,
            edge_mode=edge_mode,
            prune_isolated=prune_isolated,
        )
    )

    if fmt == "dot":
        result = export_dot(
            graph,
            rankdir=rankdir,
            cluster_depth=cluster_depth,
            cluster_mode=cluster_mode,
            style=style,
            edge_mode=edge_mode,
        )
    elif fmt == "svg":
        result = _export_svg(
            graph,
            layout=layout,
            rankdir=rankdir,
            cluster_depth=cluster_depth,
            cluster_mode=cluster_mode,
            style=style,
            edge_mode=edge_mode,
        )
    elif fmt == "html":
        result = export_html(
            graph,
            engine=layout,
            rankdir=rankdir,
            cluster_depth=cluster_depth,
            cluster_mode=cluster_mode,
            style=style,
            edge_mode=edge_mode,
        )
    else:
        cycles = detect_cycles(graph)
        result = export_json(graph, cycles=cycles)

    _write_output(result, output)


# ---------------------------------------------------------------------------
# validate
# ---------------------------------------------------------------------------


@main.command("validate")
@click.argument("path", default=".", type=click.Path(exists=True, file_okay=False))
@click.option(
    "--config",
    "-c",
    "config_path",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Path to a JSON configuration file.",
)
@click.option(
    "--output",
    "-o",
    type=click.Path(dir_okay=False, writable=True),
    default=None,
    help="Write output to a file instead of stdout.",
)
@click.option(
    "--strict",
    is_flag=True,
    default=False,
    help="Exit with non-zero status if there are any violations.",
)
@click.option(
    "--output-format",
    "output_format",
    type=click.Choice(["json", "flake8", "pylint", "github"], case_sensitive=False),
    default="json",
    show_default=True,
    help="Output format for validation results (use linter formats for CI/editor integration).",
)
def cmd_validate(
    path: str,
    config_path: Optional[str],
    output: Optional[str],
    strict: bool,
    output_format: str,
) -> None:
    """Validate Python import dependencies in PATH against configured rules.

    PATH defaults to the current working directory.
    """
    try:
        config = load_config(config_path) if config_path else default_config()
    except ConfigError as exc:
        click.echo(f"Configuration error: {exc}", err=True)
        sys.exit(1)

    graph = Analyzer(path).analyze()
    cycles = detect_cycles(graph)
    rules = _extract_rules(config)
    violations = Validator(rules).validate(graph)

    if output_format == "json":
        result = export_json(
            graph, violations=cast(list[ViolationLike], violations), cycles=cycles
        )
    else:
        result = _format_lint_output(
            violations,
            graph,
            root_path=Path(path).resolve(),
            output_format=output_format,
        )
    _write_output(result, output)

    if strict and violations:
        sys.exit(1)


# ---------------------------------------------------------------------------
# export
# ---------------------------------------------------------------------------


@main.command("export")
@click.argument("path", default=".", type=click.Path(exists=True, file_okay=False))
@click.option(
    "--normalize-hyphens/--no-normalize-hyphens",
    default=True,
    show_default=True,
    help="Normalize '-' to '_' in module names.",
)
@click.option(
    "--config",
    "-c",
    "config_path",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="Path to a JSON configuration file for highlighting violations.",
)
@click.option(
    "--format",
    "-f",
    "fmt",
    type=click.Choice(["dot", "svg", "html"], case_sensitive=False),
    default="dot",
    show_default=True,
    help="Export format.",
)
@click.option(
    "--layout",
    type=click.Choice(["dot", "sfdp", "neato", "fdp"], case_sensitive=False),
    default="dot",
    show_default=True,
    help="Graphviz layout engine for svg/html output.",
)
@click.option(
    "--style",
    type=click.Choice(
        ["default", "archi", "cruiser", "navigator", "depcruise"],
        case_sensitive=False,
    ),
    default="depcruise",
    show_default=True,
    help="Graph styling preset.",
)
@click.option(
    "--rankdir",
    type=click.Choice(["LR", "TB"], case_sensitive=False),
    default="LR",
    show_default=True,
    help="Graph direction for dot layouts.",
)
@click.option(
    "--include",
    multiple=True,
    help="Regex to include module names (repeatable).",
)
@click.option(
    "--exclude",
    multiple=True,
    help="Regex to exclude module names (repeatable).",
)
@click.option(
    "--include-path",
    multiple=True,
    help="Regex to include file paths (repeatable).",
)
@click.option(
    "--exclude-path",
    multiple=True,
    help="Regex to exclude file paths (repeatable).",
)
@click.option(
    "--exclude-common-noise-paths/--no-exclude-common-noise-paths",
    default=False,
    show_default=True,
    help=(
        "Exclude common noise paths "
        "(tests/scripts/examples/stress/migrations) from graph generation."
    ),
)
@click.option(
    "--focus",
    multiple=True,
    help="Regex to focus on module names and neighbors (repeatable).",
)
@click.option(
    "--focus-depth",
    type=int,
    default=1,
    show_default=True,
    help="Neighbor depth for focus selection.",
)
@click.option(
    "--collapse-depth",
    type=int,
    default=0,
    show_default=True,
    help="Collapse modules to a package depth (0 disables).",
)
@click.option(
    "--cluster-depth",
    type=int,
    default=5,
    show_default=True,
    help="Cluster modules by package depth (0 disables).",
)
@click.option(
    "--cluster-mode",
    type=click.Choice(["path", "module"], case_sensitive=False),
    default="path",
    show_default=True,
    help="Cluster by filesystem path or module name.",
)
@click.option(
    "--aggregate-depth",
    type=int,
    default=8,
    show_default=True,
    help="Aggregate modules by path depth (0 disables).",
)
@click.option(
    "--leaf-pattern",
    multiple=True,
    help="Regex for leaf filenames to keep as nodes (repeatable).",
)
@click.option(
    "--edge-mode",
    type=click.Choice(["node", "cluster", "auto"], case_sensitive=False),
    default="auto",
    show_default=True,
    help="Edge rendering mode for graphs.",
)
@click.option(
    "--prune-isolated/--keep-isolated",
    default=False,
    show_default=True,
    help="Drop modules without incoming/outgoing edges after filtering.",
)
@click.option(
    "--include-db-connectors/--no-include-db-connectors",
    default=False,
    show_default=True,
    help="Include DB connector imports as external nodes.",
)
@click.option(
    "--include-http-hosts/--no-include-http-hosts",
    default=False,
    show_default=True,
    help="Include HTTP request hosts as external nodes.",
)
@click.option(
    "--output",
    "-o",
    type=click.Path(dir_okay=False, writable=True),
    default=None,
    help="Write output to a file instead of stdout.",
)
def cmd_export(
    path: str,
    fmt: str,
    normalize_hyphens: bool,
    config_path: Optional[str],
    layout: str,
    style: str,
    rankdir: str,
    include: tuple[str, ...],
    exclude: tuple[str, ...],
    include_path: tuple[str, ...],
    exclude_path: tuple[str, ...],
    exclude_common_noise_paths: bool,
    focus: tuple[str, ...],
    focus_depth: int,
    collapse_depth: int,
    cluster_depth: int,
    cluster_mode: str,
    aggregate_depth: int,
    leaf_pattern: tuple[str, ...],
    edge_mode: str,
    prune_isolated: bool,
    include_db_connectors: bool,
    include_http_hosts: bool,
    output: Optional[str],
) -> None:
    """Export the dependency graph of PATH to the specified format.

    PATH defaults to the current working directory.  Use ``analyze --format dot``
    for quick one-step output; this command is for explicit graph exports.
    """
    effective_exclude_paths = _effective_exclude_paths(
        exclude_path,
        exclude_common_noise_paths,
    )
    graph = Analyzer(
        path,
        normalize_hyphens=normalize_hyphens,
        include_paths=list(include_path),
        exclude_paths=effective_exclude_paths,
        include_external_patterns=_external_patterns_for_db(include_db_connectors),
        include_http_hosts=include_http_hosts,
    ).analyze()
    graph, layout, rankdir, cluster_depth, cluster_mode, style, edge_mode = (
        _apply_graph_options(
            graph,
            include=list(include),
            exclude=list(exclude),
            include_paths=list(include_path),
            exclude_paths=effective_exclude_paths,
            focus=list(focus),
            focus_depth=focus_depth,
            collapse_depth=collapse_depth,
            cluster_depth=cluster_depth,
            cluster_mode=cluster_mode,
            aggregate_depth=aggregate_depth,
            leaf_patterns=list(leaf_pattern),
            layout=layout,
            rankdir=rankdir,
            style=style,
            edge_mode=edge_mode,
            prune_isolated=prune_isolated,
        )
    )
    violations = _load_violations(config_path, graph)
    if fmt == "dot":
        result = export_dot(
            graph,
            violations=cast(list[ViolationLike], violations),
            rankdir=rankdir,
            cluster_depth=cluster_depth,
            cluster_mode=cluster_mode,
            style=style,
            edge_mode=edge_mode,
        )
    elif fmt == "svg":
        result = _export_svg(
            graph,
            cast(list[ViolationLike], violations),
            layout=layout,
            rankdir=rankdir,
            cluster_depth=cluster_depth,
            cluster_mode=cluster_mode,
            style=style,
            edge_mode=edge_mode,
        )
    else:
        result = export_html(
            graph,
            violations=cast(list[ViolationLike], violations),
            engine=layout,
            rankdir=rankdir,
            cluster_depth=cluster_depth,
            cluster_mode=cluster_mode,
            style=style,
            edge_mode=edge_mode,
        )
    _write_output(result, output)


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _write_output(content: str, output: Optional[str]) -> None:
    if content and not content.endswith("\n"):
        content += "\n"
    if output:
        output_path = Path(output).resolve()
        output_path.write_text(content, encoding="utf-8")
        click.echo(f"Output written to {output_path}", err=True)
    else:
        click.echo(content)


def _effective_exclude_paths(
    exclude_path: tuple[str, ...],
    exclude_common_noise_paths: bool,
) -> list[str]:
    effective = list(exclude_path)
    if not exclude_common_noise_paths:
        return effective
    for pattern in DEFAULT_NOISE_PATH_PATTERNS:
        if pattern not in effective:
            effective.append(pattern)
    return effective


def _external_patterns_for_db(
    include_db_connectors: bool,
) -> list[str]:
    if not include_db_connectors:
        return []
    return list(DB_CONNECTOR_IMPORT_PATTERNS)


def _apply_graph_options(
    graph,
    include: list[str],
    exclude: list[str],
    include_paths: list[str],
    exclude_paths: list[str],
    focus: list[str],
    focus_depth: int,
    collapse_depth: int,
    cluster_depth: int,
    cluster_mode: str,
    aggregate_depth: int,
    leaf_patterns: list[str],
    layout: str,
    rankdir: str,
    style: str,
    edge_mode: str,
    prune_isolated: bool,
):
    filtered = filter_graph(
        graph,
        include=include or None,
        exclude=exclude or None,
        include_paths=include_paths or None,
        exclude_paths=exclude_paths or None,
        focus=focus or None,
        focus_depth=focus_depth,
    )
    filtered = _drop_dangling_init_modules(filtered)
    if style in {"archi", "navigator"}:
        layout = "dot"
        rankdir = "TB"
        cluster_mode = "path"
        if aggregate_depth == 0:
            aggregate_depth = 20
        if not leaf_patterns:
            leaf_patterns = [
                r"__init__\.py",
                r"models?\.py",
                r"schema\.py",
                r"types?\.py",
                r"dto\.py",
                r"api\.py",
                r"service\.py",
                r"repository\.py",
                r"client\.py",
                r"config\.py",
                r"settings?\.py",
            ]
        if cluster_depth == 0:
            cluster_depth = aggregate_depth
        compiled_leaf = [re.compile(p) for p in leaf_patterns]
        aggregated = aggregate_by_path(
            filtered,
            aggregate_depth,
            leaf_patterns=compiled_leaf,
        )
        if len(aggregated.modules) < 4:
            expanded = leaf_patterns + [r"\.py$"]
            aggregated = aggregate_by_path(
                filtered,
                aggregate_depth,
                leaf_patterns=[re.compile(p) for p in expanded],
            )
        if aggregated.modules:
            filtered = aggregated
        collapse_depth = 0
        if edge_mode == "auto":
            edge_mode = "cluster"
    if style == "cruiser":
        layout = "dot"
        rankdir = "TB"
        cluster_mode = "path"
        if cluster_depth == 3:
            cluster_depth = 5
        if edge_mode == "auto":
            edge_mode = "node"
    if edge_mode == "auto":
        edge_mode = "node"

    collapsed = collapse_graph(filtered, collapse_depth)
    pruned = prune_orphan_init_modules(collapsed)
    if prune_isolated:
        pruned = prune_isolated_modules(pruned)
    return pruned, layout, rankdir, cluster_depth, cluster_mode, style, edge_mode


def _drop_dangling_init_modules(graph: DependencyGraph) -> DependencyGraph:
    connected: set[str] = set()
    for dep in graph.dependencies:
        connected.add(dep.source)
        connected.add(dep.target)

    dangling_init = {
        module.name
        for module in graph.modules
        if module.path.endswith("__init__.py") and module.name not in connected
    }
    if not dangling_init:
        return graph

    trimmed = DependencyGraph()
    for module in graph.modules:
        if module.name in dangling_init:
            continue
        trimmed.add_module(module)

    keep_names = {module.name for module in trimmed.modules}
    for dep in graph.dependencies:
        if dep.source in keep_names and dep.target in keep_names:
            trimmed.add_dependency(dep)
    return trimmed


def _export_svg(
    graph: DependencyGraph,
    violations: list[ViolationLike] | None = None,
    layout: str = "dot",
    rankdir: str = "LR",
    cluster_depth: int = 3,
    cluster_mode: str = "path",
    style: str = "depcruise",
    edge_mode: str = "node",
) -> str:
    if violations is None:
        violations = []
    try:
        return export_svg(
            graph,
            violations=violations,
            engine=layout,
            rankdir=rankdir,
            cluster_depth=cluster_depth,
            cluster_mode=cluster_mode,
            style=style,
            edge_mode=edge_mode,
        )
    except RuntimeError as exc:
        click.echo(f"Graphviz error: {exc}", err=True)
        sys.exit(2)


def _load_violations(
    config_path: Optional[str], graph: DependencyGraph
) -> list[Violation]:
    if not config_path:
        return []
    try:
        config = load_config(config_path)
    except ConfigError as exc:
        click.echo(f"Configuration error: {exc}", err=True)
        sys.exit(1)
    rules = _extract_rules(config)
    return Validator(rules).validate(graph)


def _extract_rules(config: JSONDict) -> list[JSONDict]:
    rules_raw = config.get("rules", [])
    if not isinstance(rules_raw, list):
        return []
    return [rule for rule in rules_raw if isinstance(rule, dict)]


def _format_lint_output(
    violations: list[Violation],
    graph: DependencyGraph,
    root_path: Path,
    output_format: str,
) -> str:
    if not violations:
        return ""

    dep_lines = {(d.source, d.target): d.line for d in graph.dependencies}
    lines: list[str] = []

    for violation in violations:
        module = graph.get_module(violation.source)
        raw_path = module.path if module else violation.source.replace(".", "/") + ".py"
        rel_path = _to_display_path(raw_path, root_path)
        line_no = dep_lines.get((violation.source, violation.target), 1) or 1
        col_no = 1
        code = _lint_code(violation.severity)
        message = f"{violation.message} [rule: {violation.rule_name}]"

        if output_format == "flake8":
            lines.append(f"{rel_path}:{line_no}:{col_no}: {code} {message}")
        elif output_format == "pylint":
            lines.append(f"{rel_path}:{line_no}: [{code}] {message}")
        else:
            level = _github_level(violation.severity)
            escaped = (
                message.replace("%", "%25").replace("\n", "%0A").replace("\r", "%0D")
            )
            lines.append(
                f"::{level} file={rel_path},line={line_no},col={col_no}::{escaped}"
            )

    return "\n".join(lines)


def _to_display_path(path_value: str, root_path: Path) -> str:
    raw = Path(path_value)
    try:
        return str(raw.resolve().relative_to(root_path)).replace("\\", "/")
    except ValueError:
        return str(raw).replace("\\", "/")


def _lint_code(severity: str) -> str:
    return {
        "error": "IC001",
        "warn": "IC002",
        "info": "IC003",
    }.get(severity, "IC001")


def _github_level(severity: str) -> str:
    return {
        "error": "error",
        "warn": "warning",
        "info": "notice",
    }.get(severity, "error")


if __name__ == "__main__":
    main()



"""Dependency graph data structure for import_cruiser."""
# pylint: disable=duplicate-code

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
import os
from pathlib import Path
import re
from typing import Iterable


@dataclass
class Module:
    """Represents a Python module discovered in the project."""

    name: str
    path: str
    imports: list[str] = field(default_factory=list)

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Module):
            return NotImplemented
        return self.name == other.name


@dataclass
class Dependency:
    """Represents a directed dependency edge between two modules."""

    source: str
    target: str
    line: int = 0

    def __hash__(self) -> int:
        return hash((self.source, self.target))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Dependency):
            return NotImplemented
        return self.source == other.source and self.target == other.target


class DependencyGraph:
    """Directed graph of module dependencies."""

    def __init__(self) -> None:
        self._modules: dict[str, Module] = {}
        self._dependencies: list[Dependency] = []
        self._dependency_pairs: set[tuple[str, str]] = set()

    # ------------------------------------------------------------------
    # Mutation
    # ------------------------------------------------------------------

    def add_module(self, module: Module) -> None:
        """Register a module; silently replaces existing entry with same name."""
        self._modules[module.name] = module

    def add_dependency(self, dep: Dependency) -> None:
        """Add a dependency edge (duplicate source→target pairs are ignored)."""
        key = (dep.source, dep.target)
        if key not in self._dependency_pairs:
            self._dependencies.append(dep)
            self._dependency_pairs.add(key)

    # ------------------------------------------------------------------
    # Queries
    # ------------------------------------------------------------------

    @property
    def modules(self) -> list[Module]:
        return list(self._modules.values())

    @property
    def dependencies(self) -> list[Dependency]:
        return list(self._dependencies)

    def get_module(self, name: str) -> Module | None:
        return self._modules.get(name)

    def dependents_of(self, module_name: str) -> list[str]:
        """Return all modules that *import* ``module_name``."""
        return [d.source for d in self._dependencies if d.target == module_name]

    def dependencies_of(self, module_name: str) -> list[str]:
        """Return all modules that ``module_name`` imports."""
        return [d.target for d in self._dependencies if d.source == module_name]

    def edges(self) -> Iterable[tuple[str, str]]:
        for dep in self._dependencies:
            yield dep.source, dep.target


def detect_cycles(graph: DependencyGraph) -> list[list[str]]:
    """Return a list of cycles found in *graph*."""
    adj: dict[str, list[str]] = defaultdict(list)
    for source, target in graph.edges():
        adj[source].append(target)

    all_modules = {m.name for m in graph.modules}

    visited: set[str] = set()
    rec_stack: set[str] = set()
    cycles: list[list[str]] = []
    seen_cycles: set[frozenset[str]] = set()

    def dfs(node: str, path: list[str]) -> None:
        visited.add(node)
        rec_stack.add(node)
        path.append(node)

        for neighbour in adj.get(node, []):
            if neighbour not in all_modules:
                continue
            if neighbour not in visited:
                dfs(neighbour, path)
            elif neighbour in rec_stack:
                cycle_start = path.index(neighbour)
                cycle = path[cycle_start:]
                normalised = frozenset(cycle)
                if normalised not in seen_cycles:
                    cycles.append(list(cycle))
                    seen_cycles.add(normalised)

        path.pop()
        rec_stack.discard(node)

    for module in sorted(all_modules):
        if module not in visited:
            dfs(module, [])

    return cycles


def filter_graph(
    graph: DependencyGraph,
    include: list[str] | None = None,
    exclude: list[str] | None = None,
    include_paths: list[str] | None = None,
    exclude_paths: list[str] | None = None,
    focus: list[str] | None = None,
    focus_depth: int = 1,
) -> DependencyGraph:
    include_patterns = _compile_patterns(include)
    exclude_patterns = _compile_patterns(exclude)
    include_path_patterns = _compile_patterns(include_paths)
    exclude_path_patterns = _compile_patterns(exclude_paths)
    focus_patterns = _compile_patterns(focus)

    module_map = {m.name: m for m in graph.modules}
    root_path = _common_root(graph.modules)
    allowed = set(module_map.keys())

    allowed = _apply_name_filters(allowed, include_patterns, exclude_patterns)
    allowed = _apply_include_path_filters(
        graph,
        module_map,
        allowed,
        root_path,
        include_path_patterns,
    )
    allowed = _apply_exclude_path_filters(
        module_map,
        allowed,
        root_path,
        exclude_path_patterns,
    )

    if focus_patterns:
        focus_set = {name for name in allowed if _matches_any(name, focus_patterns)}
        allowed = _expand_focus(graph, focus_set, allowed, focus_depth)

    return _subgraph(graph, module_map, allowed)


def collapse_graph(graph: DependencyGraph, depth: int) -> DependencyGraph:
    if depth <= 0:
        return graph

    collapsed = DependencyGraph()
    name_map: dict[str, str] = {}
    for module in graph.modules:
        collapsed_name = _collapse_name(module.name, depth)
        name_map[module.name] = collapsed_name
        if collapsed.get_module(collapsed_name) is None:
            collapsed.add_module(Module(name=collapsed_name, path=module.path))

    for dep in graph.dependencies:
        src = name_map.get(dep.source)
        tgt = name_map.get(dep.target)
        if src and tgt and src != tgt:
            collapsed.add_dependency(Dependency(source=src, target=tgt, line=dep.line))

    return collapsed


def prune_orphan_init_modules(graph: DependencyGraph) -> DependencyGraph:
    """Remove modules backed by ``__init__.py`` files that have no edges."""
    module_map = {m.name: m for m in graph.modules}
    connected: set[str] = set()
    for dep in graph.dependencies:
        connected.add(dep.source)
        connected.add(dep.target)

    allowed = {
        name
        for name, module in module_map.items()
        if Path(module.path).name != "__init__.py" or name in connected
    }
    return _subgraph(graph, module_map, allowed)


def prune_isolated_modules(graph: DependencyGraph) -> DependencyGraph:
    """Remove modules that have no incoming or outgoing edges."""
    module_map = {m.name: m for m in graph.modules}
    connected: set[str] = set()
    for dep in graph.dependencies:
        connected.add(dep.source)
        connected.add(dep.target)

    allowed = {name for name in module_map if name in connected}
    return _subgraph(graph, module_map, allowed)


def aggregate_by_path(
    graph: DependencyGraph,
    depth: int,
    leaf_patterns: list[re.Pattern[str]] | None = None,
) -> DependencyGraph:
    if depth <= 0:
        return graph

    root = _common_root(graph.modules)
    if root is None:
        return graph

    init_files = {
        _dir_key(Path(module.path).resolve().parent)
        for module in graph.modules
        if module.path.endswith("__init__.py")
    }

    aggregated = DependencyGraph()
    group_map: dict[str, str] = {}
    group_path_map: dict[str, str] = {}

    for module in graph.modules:
        group_key = _group_key(module.path, root, depth)
        if not group_key:
            continue
        filename = Path(module.path).name
        if leaf_patterns and any(p.search(filename) for p in leaf_patterns):
            leaf_key = os.path.join(group_key, filename)
            group_map[module.name] = leaf_key
            group_path_map.setdefault(leaf_key, module.path)
        else:
            group_map[module.name] = group_key
            if group_key not in group_path_map:
                group_dir = Path(root) / Path(group_key)
                init_path = group_dir / "__init__.py"
                path_key = _dir_key(group_dir)
                if path_key not in init_files:
                    group_path_map[group_key] = str(init_path)
                else:
                    group_path_map[group_key] = str(init_path)

    for group_key, path in group_path_map.items():
        aggregated.add_module(Module(name=group_key, path=path))

    for dep in graph.dependencies:
        src = group_map.get(dep.source)
        tgt = group_map.get(dep.target)
        if src and tgt and src != tgt:
            aggregated.add_dependency(Dependency(source=src, target=tgt, line=dep.line))

    return aggregated


def _compile_patterns(patterns: list[str] | None) -> list[re.Pattern[str]]:
    if not patterns:
        return []
    return [re.compile(p) for p in patterns]


def _matches_any(value: str, patterns: list[re.Pattern[str]]) -> bool:
    return any(p.search(value) for p in patterns)


def _matches_path_filters(
    path: str,
    root_path: str | None,
    patterns: list[re.Pattern[str]],
) -> bool:
    rel_path = _match_path(path, root_path)
    abs_path = path.replace("\\", "/")
    return _matches_any(rel_path, patterns) or _matches_any(abs_path, patterns)


def _apply_name_filters(
    allowed: set[str],
    include_patterns: list[re.Pattern[str]],
    exclude_patterns: list[re.Pattern[str]],
) -> set[str]:
    filtered = allowed
    if include_patterns:
        filtered = {name for name in filtered if _matches_any(name, include_patterns)}
    if exclude_patterns:
        filtered = {
            name for name in filtered if not _matches_any(name, exclude_patterns)
        }
    return filtered


def _apply_include_path_filters(
    graph: DependencyGraph,
    module_map: dict[str, Module],
    allowed: set[str],
    root_path: str | None,
    include_path_patterns: list[re.Pattern[str]],
) -> set[str]:
    if not include_path_patterns:
        return allowed
    allowed_with_path = {
        name
        for name in allowed
        if module_map[name].path
        and _matches_path_filters(
            module_map[name].path,
            root_path,
            include_path_patterns,
        )
    }
    pathless = {name for name in allowed if not module_map[name].path}
    connected_pathless = {
        dep.target
        for dep in graph.dependencies
        if dep.source in allowed_with_path and dep.target in pathless
    }
    connected_pathless.update(
        dep.source
        for dep in graph.dependencies
        if dep.target in allowed_with_path and dep.source in pathless
    )
    return allowed_with_path | connected_pathless


def _apply_exclude_path_filters(
    module_map: dict[str, Module],
    allowed: set[str],
    root_path: str | None,
    exclude_path_patterns: list[re.Pattern[str]],
) -> set[str]:
    if not exclude_path_patterns:
        return allowed
    return {
        name
        for name in allowed
        if not _matches_path_filters(
            module_map[name].path,
            root_path,
            exclude_path_patterns,
        )
    }


def _expand_focus(
    graph: DependencyGraph,
    focus_set: set[str],
    allowed: set[str],
    depth: int,
) -> set[str]:
    if not focus_set:
        return set()
    if depth <= 0:
        return focus_set

    adj_out: dict[str, set[str]] = {}
    adj_in: dict[str, set[str]] = {}
    for dep in graph.dependencies:
        if dep.source in allowed and dep.target in allowed:
            adj_out.setdefault(dep.source, set()).add(dep.target)
            adj_in.setdefault(dep.target, set()).add(dep.source)

    visited = set(focus_set)
    frontier = set(focus_set)
    for _ in range(depth):
        next_frontier: set[str] = set()
        for node in frontier:
            next_frontier.update(adj_out.get(node, set()))
            next_frontier.update(adj_in.get(node, set()))
        next_frontier = {n for n in next_frontier if n in allowed}
        new_nodes = next_frontier - visited
        if not new_nodes:
            break
        visited.update(new_nodes)
        frontier = new_nodes

    return visited


def _subgraph(
    graph: DependencyGraph,
    module_map: dict[str, Module],
    allowed: set[str],
) -> DependencyGraph:
    filtered = DependencyGraph()
    for name in sorted(allowed):
        module = module_map.get(name)
        if module:
            filtered.add_module(Module(name=module.name, path=module.path))

    for dep in graph.dependencies:
        if dep.source in allowed and dep.target in allowed:
            filtered.add_dependency(
                Dependency(source=dep.source, target=dep.target, line=dep.line)
            )

    return filtered


def _collapse_name(name: str, depth: int) -> str:
    parts = name.split(".")
    if len(parts) <= depth:
        return name
    return ".".join(parts[:depth])


def _match_path(path: str, root_path: str | None) -> str:
    if not path:
        return ""
    if root_path:
        try:
            rel = os.path.relpath(path, root_path)
        except ValueError:
            rel = path
    else:
        rel = path
    return rel.replace(os.sep, "/")


def _common_root(modules: list[Module]) -> str | None:
    paths = [m.path for m in modules if m.path]
    if not paths:
        return None
    try:
        return os.path.commonpath(paths)
    except ValueError:
        return None


def _group_key(path: str, root: str, depth: int) -> str:
    root_path = Path(root).resolve()
    rel = Path(path).resolve().relative_to(root_path)
    parts = list(rel.parts[:-1])  # directories only
    if not parts:
        return root_path.name
    return os.path.join(*parts[: min(depth, len(parts))])


def _dir_key(path: Path) -> str:
    return str(path.resolve())



"""import-cruiser – Analyze, validate, and visualize Python import dependencies."""

from .analyzer import Analyzer
from .config import ConfigError, default_config, load_config, validate_config
from .detector import detect_cycles
from .exporter import (
    export_dot,
    export_html,
    export_json,
    export_svg,
)
from .graph import (
    Dependency,
    DependencyGraph,
    Module,
    aggregate_by_path,
    collapse_graph,
    filter_graph,
)
from .validator import Validator, Violation

__all__ = [
    "Analyzer",
    "ConfigError",
    "Dependency",
    "DependencyGraph",
    "Module",
    "Validator",
    "Violation",
    "aggregate_by_path",
    "collapse_graph",
    "default_config",
    "detect_cycles",
    "export_dot",
    "export_html",
    "export_json",
    "export_svg",
    "filter_graph",
    "load_config",
    "validate_config",
]

__version__ = "0.2.6"
